import sys

from cbapi.example_helpers import build_cli_parser, get_cb_livequery_object
from cbapi.psc.livequery.models import Run


def create_run(cb, args):
    query = cb.query(args.sql)

    if args.device_ids:
        query.device_ids(args.device_ids)
    if args.device_types:
        query.device_types(args.device_types)
    if args.policy_ids:
        query.policy_ids(args.policy_ids)
    if args.notify:
        query.notify_on_finish()
    if args.name:
        query.name(args.name)

    run = query.submit()
    print(run)


def run_status(cb, args):
    run = cb.select(Run, args.id)
    print(run)


def run_stop(cb, args):
    run = cb.select(Run, args.id)
    if run.stop():
        print("Run {} has been stopped.".format(run.id))
        print(run)
    else:
        print("Unable to stop run {}".format(run.id))


def run_delete(cb, args):
    run = cb.select(Run, args.id)
    if run.delete():
        print("Run {} has been deleted.".format(run.id))
    else:
        print("Unable to delete run {}".format(run.id))


def run_history(cb, args):
    results = cb.query_history(args.query)
    if args.sort_by:
        dir = "DESC" if args.descending_results else "ASC"
        results.sort_by(args.sort_by, direction=dir)
    for result in results:
        print(result)


def main():
    parser = build_cli_parser("Create and manage LiveQuery runs")
    commands = parser.add_subparsers(help="Commands", dest="command_name")

    create_command = commands.add_parser("create", help="Create a new LiveQuery run")
    create_command.add_argument(
        "-s", "--sql", type=str, required=True, help="The query to run"
    )
    create_command.add_argument(
        "-n",
        "--notify",
        action="store_true",
        help="Notify by email when the run finishes",
    )
    create_command.add_argument(
        "-N", "--name", type=str, required=False, help="The name of the run"
    )
    create_command.add_argument(
        "--device_ids",
        nargs="+",
        type=int,
        required=False,
        help="Device IDs to filter on",
    )
    create_command.add_argument(
        "--device_types",
        nargs="+",
        type=str,
        required=False,
        help="Device types to filter on",
    )
    create_command.add_argument(
        "--policy_ids",
        nargs="+",
        type=str,
        required=False,
        help="Policy IDs to filter on",
    )

    status_command = commands.add_parser(
        "status", help="Retrieve information about a run"
    )
    status_command.add_argument(
        "-i", "--id", type=str, required=True, help="The run ID"
    )

    stop_command = commands.add_parser(
        "stop", help="Stops/cancels a current run"
    )
    stop_command.add_argument(
        "-i", "--id", type=str, required=True, help="The run ID"
    )

    delete_command = commands.add_parser(
        "delete", help="Permanently delete a run"
    )
    delete_command.add_argument(
        "-i", "--id", type=str, required=True, help="The run ID"
    )

    history_command = commands.add_parser(
        "history", help="List history of all runs"
    )
    history_command.add_argument(
        "-q", "--query", type=str, required=False, help="Query string to use"
    )
    history_command.add_argument(
        "-S", "--sort_by", type=str, help="sort by this field", required=False
    )
    history_command.add_argument(
        "-D",
        "--descending_results",
        help="return results in descending order",
        action="store_true",
        required=False
    )

    args = parser.parse_args()
    cb = get_cb_livequery_object(args)

    if args.command_name == "create":
        return create_run(cb, args)
    elif args.command_name == "status":
        return run_status(cb, args)
    elif args.command_name == "stop":
        return run_stop(cb, args)
    elif args.command_name == "delete":
        return run_delete(cb, args)
    elif args.command_name == "history":
        return run_history(cb, args)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

from __future__ import absolute_import

from copy import deepcopy

from cbapi.six import python_2_unicode_compatible

import base64
import os.path
from cbapi.six import iteritems, add_metaclass, integer_types
from cbapi.six.moves import range
from .response.utils import convert_from_cb, convert_to_cb
import yaml
import json
import time
from .errors import ApiError, ServerError, InvalidObjectError
from .oldmodels import BaseModel
import logging
from datetime import datetime
from cbapi.utils import calculate_elapsed_time

log = logging.getLogger(__name__)


class CreatableModelMixin(object):
    pass


class CbMetaModel(type):
    model_base_directory = os.path.dirname(__file__)
    model_classes = []

    def __new__(mcs, name, bases, clsdict):
        swagger_meta_file = clsdict.pop("swagger_meta_file", None)
        model_data = {}
        if swagger_meta_file:
            model_data = yaml.safe_load(
                open(os.path.join(mcs.model_base_directory, swagger_meta_file), 'rb').read())

        clsdict["__doc__"] = "Represents a %s object in the Carbon Black server.\n\n" % (name,)
        for field_name, field_info in iteritems(model_data.get("properties", {})):
            docstring = field_info.get("description", None)
            if docstring:
                clsdict["__doc__"] += ":ivar %s: %s\n" % (field_name, docstring)

        foreign_keys = clsdict.pop("foreign_keys", {})

        cls = super(CbMetaModel, mcs).__new__(mcs, name, bases, clsdict)
        mcs.model_classes.append(cls)

        cls._valid_fields = []
        cls._required_fields = model_data.get("required", [])
        cls._default_value = {}

        for field_name, field_info in iteritems(model_data.get("properties", {})):
            cls._valid_fields.append(field_name)

            default_value = field_info.get("default", None)
            if default_value:
                cls._default_value[field_name] = default_value

            field_format = field_info.get("type", "string")
            field_format = field_info.get("format", field_format)

            if field_format.startswith('int'):
                setattr(cls, field_name, FieldDescriptor(field_name, coerce_to=int))
            elif field_format == "iso-date-time":
                setattr(cls, field_name, IsoDateTimeFieldDescriptor(field_name))
            elif field_format == "epoch-ms-date-time":
                setattr(cls, field_name, EpochDateTimeFieldDescriptor(field_name, 1000.0))
            elif field_format == "boolean":
                setattr(cls, field_name, FieldDescriptor(field_name, coerce_to=bool))
            elif field_format == "array":
                setattr(cls, field_name, ArrayFieldDescriptor(field_name))
            elif field_format == "object":
                setattr(cls, field_name, ObjectFieldDescriptor(field_name))
            elif field_format == "double":
                setattr(cls, field_name, FieldDescriptor(field_name, coerce_to=float))
            elif field_format == "byte":
                setattr(cls, field_name, BinaryFieldDescriptor(field_name))
            else:
                setattr(cls, field_name, FieldDescriptor(field_name))

        for fk_name, fk_info in iteritems(foreign_keys):
            setattr(cls, fk_name, ForeignKeyFieldDescriptor(fk_name, fk_info[0], fk_info[1]))

        return cls


class FieldDescriptor(object):
    def __init__(self, field_name, coerce_to=None, default_value=None):
        self.att_name = field_name
        self.default_value = default_value
        self.coerce_to = coerce_to

    def __get__(self, instance, instance_type=None):
        if instance is not None:
            if self.att_name not in instance._info and not instance._full_init:
                instance.refresh()

            value = instance._info.get(self.att_name, self.default_value)
            coerce_type = self.coerce_to or type(value)
            if value is None:
                return None
            return coerce_type(value)

    def __set__(self, instance, value):
        coerce_type = self.coerce_to or type(value)
        value = coerce_type(value)
        instance._set(self.att_name, value)


class ArrayFieldDescriptor(FieldDescriptor):
    def __get__(self, instance, instance_type=None):
        ret = super(ArrayFieldDescriptor, self).__get__(instance, instance_type)
        return ret or []


# TODO: this is a kludge to avoid writing "small" models?
class ObjectFieldDescriptor(FieldDescriptor):
    def __get__(self, instance, instance_type=None):
        ret = super(ObjectFieldDescriptor, self).__get__(instance, instance_type)
        return ret or {}


class IsoDateTimeFieldDescriptor(FieldDescriptor):
    def __init__(self, field_name):
        super(IsoDateTimeFieldDescriptor, self).__init__(field_name)

    def __get__(self, instance, instance_type=None):
        d = super(IsoDateTimeFieldDescriptor, self).__get__(instance, instance_type)
        return convert_from_cb(d)

    def __set__(self, instance, value):
        parsed_date = convert_to_cb(value)
        super(IsoDateTimeFieldDescriptor, self).__set__(instance, parsed_date)


class EpochDateTimeFieldDescriptor(FieldDescriptor):
    def __init__(self, field_name, multiplier=1.0):
        super(EpochDateTimeFieldDescriptor, self).__init__(field_name)
        self.multiplier = float(multiplier)

    def __get__(self, instance, instance_type=None):
        d = super(EpochDateTimeFieldDescriptor, self).__get__(instance, instance_type)
        if type(d) is float or isinstance(type(d), integer_types):
            epoch_seconds = d / self.multiplier
            return datetime.utcfromtimestamp(epoch_seconds)
        else:
            return datetime.utcfromtimestamp(0)      # default to epoch time (1970-01-01) if we have a non-numeric type

    def __set__(self, instance, value):
        if isinstance(value, datetime):
            new_value = calculate_elapsed_time(value - self.epoch) * self.multiplier
        else:
            new_value = value
        super(EpochDateTimeFieldDescriptor, self).__set__(instance, new_value)


class ForeignKeyFieldDescriptor(FieldDescriptor):
    def __init__(self, field_name, join_model, join_field=None):
        super(ForeignKeyFieldDescriptor, self).__init__(field_name)
        self.join_model = join_model
        if join_field is None:
            self.join_field = field_name + "_id"
        else:
            self.join_field = join_field

    def __get__(self, instance, instance_type=None):
        foreign_id = getattr(instance, self.join_field)
        return instance._cb.select(self.join_model, foreign_id)

    def __set__(self, instance, value):
        if type(value, BaseModel) or type(value, NewBaseModel):
            setattr(self, self.join_field, getattr(value, "_model_unique_id"))
        else:
            setattr(self, self.join_field, value)


class BinaryFieldDescriptor(FieldDescriptor):
    def __get__(self, instance, instance_type=None):
        d = super(BinaryFieldDescriptor, self).__get__(instance, instance_type)
        return base64.b64decode(d)

    def __set__(self, instance, value):
        super(BinaryFieldDescriptor, self).__set__(instance, base64.b64encode(value))


@python_2_unicode_compatible
@add_metaclass(CbMetaModel)
class NewBaseModel(object):
    primary_key = "id"

    def __init__(self, cb, model_unique_id=None, initial_data=None, force_init=False, full_doc=False):
        """
        Base model for
        :param cb:
        :param model_unique_id:
        :param initial_data:
        :param force_init:
        :param full_doc:
        :return:
        """
        self._cb = cb
        self._last_refresh_time = 0

        if initial_data is not None:
            self._info = initial_data
        else:
            self._info = {}

        if model_unique_id is not None:
            self._info[self.__class__.primary_key] = model_unique_id

        self._dirty_attributes = {}
        self._full_init = full_doc

        if force_init:
            self.refresh()

    @property
    def _model_unique_id(self):
        return self._info.get(self.__class__.primary_key, None)

    @classmethod
    def new_object(cls, cb, item, **kwargs):
        return cb.select(cls, item[cls.primary_key], initial_data=item, **kwargs)

    def __getattr__(self, item):
        try:
            super(NewBaseModel, self).__getattribute__(item)
        except AttributeError:
            pass         # fall through to the rest of the logic...

        # try looking up via self._info, if we already have it.
        if item in self._info:
            return self._info[item]

        # if we're still here, let's load the object if we haven't done so already.
        if not self._full_init:
            self._refresh()

        # try one more time.
        if item in self._info:
            return self._info[item]
        else:
            raise AttributeError("'{0}' object has no attribute '{1}'".format(self.__class__.__name__,
                                                                              item))

    def __setattr__(self, attrname, val):
        if attrname.startswith("_"):
            super(NewBaseModel, self).__setattr__(attrname, val)
        else:
            raise AttributeError("Field {0:s} is immutable".format(attrname))

    def get(self, attrname, default_val=None):
        return getattr(self, attrname, default_val)

    def _set(self, attrname, new_value):
        pass

    def refresh(self):
        return self._refresh()

    def _refresh(self):
        if self._model_unique_id is not None and self.__class__.primary_key not in self._dirty_attributes.keys():
            self._info = self._parse(self._retrieve_cb_info())

            self._full_init = True
            self._last_refresh_time = time.time()
            return True
        return False

    def _build_api_request_uri(self, http_method="GET"):
        baseuri = self.__class__.__dict__.get('urlobject', None)
        if self._model_unique_id is not None:
            return baseuri + "/%s" % self._model_unique_id
        else:
            return baseuri

    def _retrieve_cb_info(self):
        request_uri = self._build_api_request_uri()
        return self._cb.get_object(request_uri)

    def _parse(self, obj):
        return obj

    @property
    def original_document(self):
        if not self._full_init:
            self.refresh()

        return self._info

    def __repr__(self):
        if self._model_unique_id is not None:
            return "<%s.%s: id %s> @ %s" % (self.__class__.__module__, self.__class__.__name__, self._model_unique_id,
                                            self._cb.session.server)
        else:
            return "<%s.%s object at %s> @ %s" % (self.__class__.__module__, self.__class__.__name__, hex(id(self)),
                                                  self._cb.session.server)

    def __str__(self):
        lines = []
        lines.append("{0:s} object, bound to {1:s}.".format(self.__class__.__name__, self._cb.session.server))
        if self._last_refresh_time:
            lines.append(" Last refreshed at {0:s}".format(time.ctime(self._last_refresh_time)))
        if not self._full_init:
            lines.append(" Partially initialized. Use .refresh() to load all attributes")
        lines.append("-"*79)
        lines.append("")

        for attr in sorted(self._info):
            status = "   "
            if attr in self._dirty_attributes:
                if self._dirty_attributes[attr] is None:
                    status = "(+)"
                else:
                    status = "(*)"
            try:
                val = str(self._info[attr])
            except UnicodeDecodeError:
                val = repr(self._info[attr])
            if len(val) > 50:
                val = val[:47] + u"..."
            lines.append(u"{0:s} {1:>20s}: {2:s}".format(status, attr, val))

        return "\n".join(lines)

    def _join(self, join_cls, field_name):
        try:
            field_value = getattr(self, field_name)
        except AttributeError:
            return None

        if field_value is None:
            return None

        return self._cb.select(join_cls, field_value)


class UnrefreshableModel(NewBaseModel):
    """Represents a model that can't be refreshed, i.e. for which ``reset()``
    is not a valid operation.
    """
    def refresh(self):
        raise ApiError("refresh() called on an unrefreshable model")


class MutableBaseModel(NewBaseModel):
    _new_object_http_method = "POST"
    _change_object_http_method = "PUT"
    _new_object_needs_primary_key = False

    def __setattr__(self, attrname, val):
        # allow subclasses to define their own property setters
        propobj = getattr(self.__class__, attrname, None)
        if isinstance(propobj, property) and propobj.fset:
            return propobj.fset(self, val)

        if not attrname.startswith("_") and attrname not in self.__class__._valid_fields:
            if attrname in self._info:
                log.warning("Changing field not included in Swagger definition: {0:s}".format(attrname))
                self._set(attrname, val)
            else:
                print("Trying to set attribute {0:s}".format(attrname))
        else:
            object.__setattr__(self, attrname, val)

    def _set(self, attrname, new_value):
        # ensure that we are operating on the full object first
        if not self._full_init and self._model_unique_id is not None:
            self.refresh()

        # extract unique ID if we're updating a "joined" field
        if isinstance(new_value, NewBaseModel):
            new_value = new_value._model_unique_id

        # early exit if we attempt to set the field to its own value
        if new_value == self._info.get(attrname, None):
            return

        # update dirty_attributes if necessary
        if attrname in self._dirty_attributes:
            if new_value == self._dirty_attributes[attrname]:
                del self._dirty_attributes[attrname]
        else:
            self._dirty_attributes[attrname] = self._info.get(attrname, None)

        # finally, make the change
        self._info[attrname] = new_value

    def refresh(self):
        if self._refresh():
            self._dirty_attributes = {}

    def is_dirty(self):
        return len(self._dirty_attributes) > 0

    def _update_object(self):
        if self.__class__.primary_key in self._dirty_attributes.keys() or self._model_unique_id is None:
            new_object_info = deepcopy(self._info)
            try:
                if not self._new_object_needs_primary_key:
                    del(new_object_info[self.__class__.primary_key])
            except Exception:
                pass
            log.debug("Creating a new {0:s} object".format(self.__class__.__name__))
            ret = self._cb.api_json_request(self.__class__._new_object_http_method, self.urlobject,
                                            data=new_object_info)
        else:
            log.debug("Updating {0:s} with unique ID {1:s}".format(self.__class__.__name__, str(self._model_unique_id)))
            http_method = self.__class__._change_object_http_method
            ret = self._cb.api_json_request(http_method, self._build_api_request_uri(http_method=http_method),
                                            data=self._info)

        return self._refresh_if_needed(ret)

    def _refresh_if_needed(self, request_ret):
        refresh_required = False

        if request_ret.status_code not in range(200, 300):
            try:
                message = json.loads(request_ret.text)[0]
            except Exception:
                message = request_ret.text

            raise ServerError(request_ret.status_code, message,
                              result="Did not update {} record.".format(self.__class__.__name__))
        else:
            try:
                message = request_ret.json()
                log.debug("Received response: %s" % message)
                if list(message.keys()) == ["result"]:
                    post_result = message.get("result", None)

                    if post_result and post_result != "success":
                        raise ServerError(request_ret.status_code, post_result,
                                          result="Did not update {0:s} record.".format(self.__class__.__name__))
                    else:
                        refresh_required = True
                else:
                    self._info = json.loads(request_ret.text)
                    if message.keys() == ["id"]:
                        # if all we got back was an ID, try refreshing to get the entire record.
                        log.debug("Only received an ID back from the server, forcing a refresh")
                        refresh_required = True
                    else:
                        self._full_init = True
            except Exception:
                refresh_required = True

        self._dirty_attributes = {}
        if refresh_required:
            self.refresh()
        return self._model_unique_id

    def save(self):
        if not self.is_dirty():
            return

        self.validate()
        self._update_object()
        return self

    def reset(self):
        for k, v in iteritems(self._dirty_attributes):
            if v is None:
                del self._info[k]
            else:
                self._info[k] = v

        self._dirty_attributes = {}

    # TODO: How do we delete this object from our LRU cache?
    def delete(self):
        return self._delete_object()

    def _delete_object(self):
        if self._model_unique_id:
            ret = self._cb.delete_object(self._build_api_request_uri())
        else:
            return

        if ret.status_code not in (200, 204):
            try:
                message = json.loads(ret.text)[0]
            except Exception:
                message = ret.text
            raise ServerError(ret.status_code, message, result="Did not delete {0:s}.".format(str(self)))

    def validate(self):
        if not self._full_init:
            self.refresh()

        diff = list(set(self.__class__._required_fields) - set(self._info.keys()))
        if not diff:
            return True
        else:
            raise InvalidObjectError("Missing fields: [%s]" % (", ".join(diff)))

    def __repr__(self):
        r = super(MutableBaseModel, self).__repr__()
        if self.is_dirty():
            r += " (*)"
        return r



from cbapi.errors import ApiError, MoreThanOneResultError
import functools
from six import string_types
from solrq import Q


class QueryBuilder(object):
    """
    Provides a flexible interface for building prepared queries for the CB
    PSC backend.

    This object can be instantiated directly, or can be managed implicitly
    through the :py:meth:`select` API.
    """

    def __init__(self, **kwargs):
        if kwargs:
            self._query = Q(**kwargs)
        else:
            self._query = None
        self._raw_query = None

    def _guard_query_params(func):
        """Decorates the query construction methods of *QueryBuilder*, preventing
        them from being called with parameters that would result in an internally
        inconsistent query.
        """

        @functools.wraps(func)
        def wrap_guard_query_change(self, q, **kwargs):
            if self._raw_query is not None and (kwargs or isinstance(q, Q)):
                raise ApiError("Cannot modify a raw query with structured parameters")
            if self._query is not None and isinstance(q, string_types):
                raise ApiError("Cannot modify a structured query with a raw parameter")
            return func(self, q, **kwargs)

        return wrap_guard_query_change

    @_guard_query_params
    def where(self, q, **kwargs):
        """Adds a conjunctive filter to a query.

        :param q: string or `solrq.Q` object
        :param kwargs: Arguments to construct a `solrq.Q` with
        :return: QueryBuilder object
        :rtype: :py:class:`QueryBuilder`
        """
        if isinstance(q, string_types):
            if self._raw_query is None:
                self._raw_query = []
            self._raw_query.append(q)
        elif isinstance(q, Q) or kwargs:
            if self._query is not None:
                raise ApiError("Use .and_() or .or_() for an extant solrq.Q object")
            if kwargs:
                q = Q(**kwargs)
            self._query = q
        else:
            raise ApiError(".where() only accepts strings or solrq.Q objects")

        return self

    @_guard_query_params
    def and_(self, q, **kwargs):
        """Adds a conjunctive filter to a query.

        :param q: string or `solrq.Q` object
        :param kwargs: Arguments to construct a `solrq.Q` with
        :return: QueryBuilder object
        :rtype: :py:class:`QueryBuilder`
        """
        if isinstance(q, string_types):
            self.where(q)
        elif isinstance(q, Q) or kwargs:
            if kwargs:
                q = Q(**kwargs)
            if self._query is None:
                self._query = q
            else:
                self._query = self._query & q
        else:
            raise ApiError(".and_() only accepts strings or solrq.Q objects")

        return self

    @_guard_query_params
    def or_(self, q, **kwargs):
        """Adds a disjunctive filter to a query.

        :param q: `solrq.Q` object
        :param kwargs: Arguments to construct a `solrq.Q` with
        :return: QueryBuilder object
        :rtype: :py:class:`QueryBuilder`
        """
        if kwargs:
            q = Q(**kwargs)

        if isinstance(q, Q):
            if self._query is None:
                self._query = q
            else:
                self._query = self._query | q
        else:
            raise ApiError(".or_() only accepts solrq.Q objects")

        return self

    @_guard_query_params
    def not_(self, q, **kwargs):
        """Adds a negative filter to a query.

        :param q: `solrq.Q` object
        :param kwargs: Arguments to construct a `solrq.Q` with
        :return: QueryBuilder object
        :rtype: :py:class:`QueryBuilder`
        """
        if kwargs:
            q = ~Q(**kwargs)

        if isinstance(q, Q):
            if self._query is None:
                self._query = q
            else:
                self._query = self._query & q
        else:
            raise ApiError(".not_() only accepts solrq.Q objects")

    def _collapse(self):
        """The query can be represented by either an array of strings
        (_raw_query) which is concatenated and passed directly to Solr, or
        a solrq.Q object (_query) which is then converted into a string to
        pass to Solr. This function will perform the appropriate conversions to
        end up with the 'q' string sent into the POST request to the
        PSC-R query endpoint."""
        if self._raw_query is not None:
            return " ".join(self._raw_query)
        elif self._query is not None:
            return str(self._query)
        else:
            return None  # return everything


class PSCQueryBase:
    """
    Represents the base of all LiveQuery query classes.
    """

    def __init__(self, doc_class, cb):
        self._doc_class = doc_class
        self._cb = cb


class QueryBuilderSupportMixin:
    """
    A mixin that supplies wrapper methods to access the _query_builder.
    """
    def where(self, q=None, **kwargs):
        """Add a filter to this query.

        :param q: Query string, :py:class:`QueryBuilder`, or `solrq.Q` object
        :param kwargs: Arguments to construct a `solrq.Q` with
        :return: Query object
        :rtype: :py:class:`Query`
        """

        if not q:
            return self
        if isinstance(q, QueryBuilder):
            self._query_builder = q
        else:
            self._query_builder.where(q, **kwargs)
        return self

    def and_(self, q=None, **kwargs):
        """Add a conjunctive filter to this query.

        :param q: Query string or `solrq.Q` object
        :param kwargs: Arguments to construct a `solrq.Q` with
        :return: Query object
        :rtype: :py:class:`Query`
        """
        if not q and not kwargs:
            raise ApiError(".and_() expects a string, a solrq.Q, or kwargs")

        self._query_builder.and_(q, **kwargs)
        return self

    def or_(self, q=None, **kwargs):
        """Add a disjunctive filter to this query.

        :param q: `solrq.Q` object
        :param kwargs: Arguments to construct a `solrq.Q` with
        :return: Query object
        :rtype: :py:class:`Query`
        """
        if not q and not kwargs:
            raise ApiError(".or_() expects a solrq.Q or kwargs")

        self._query_builder.or_(q, **kwargs)
        return self

    def not_(self, q=None, **kwargs):
        """Adds a negated filter to this query.

        :param q: `solrq.Q` object
        :param kwargs: Arguments to construct a `solrq.Q` with
        :return: Query object
        :rtype: :py:class:`Query`
        """

        if not q and not kwargs:
            raise ApiError(".not_() expects a solrq.Q, or kwargs")

        self._query_builder.not_(q, **kwargs)
        return self


class IterableQueryMixin:
    """
    A mix-in to provide iterability to a query.
    """
    def all(self):
        """
        Returns all the items of a query as a list.

        :return: List of query items
        """
        return self._perform_query()

    def first(self):
        """
        Returns the first item that would be returned as the result of a query.

        :return: First query item
        """
        allres = list(self)
        res = allres[:1]
        if not len(res):
            return None
        return res[0]

    def one(self):
        """
        Returns the only item that would be returned by a query.

        :return: Sole query return item
        :raises MoreThanOneResultError: If the query returns zero items, or more than one item
        """
        allres = list(self)
        res = allres[:2]
        if len(res) == 0:
            raise MoreThanOneResultError(
                message="0 results for query {0:s}".format(self._query)
            )
        if len(res) > 1:
            raise MoreThanOneResultError(
                message="{0:d} results found for query {1:s}".format(
                    len(self), self._query
                )
            )
        return res[0]

    def __len__(self):
        return 0

    def __getitem__(self, item):
        return None

    def __iter__(self):
        return self._perform_query()



#!/usr/bin/env python
#

import sys
from cbapi.response.models import SensorGroup, Site
from cbapi.example_helpers import build_cli_parser, get_cb_response_object
from cbapi.errors import ServerError
import logging

log = logging.getLogger(__name__)


def list_sensor_groups(cb, parser, args):
    for g in cb.select(SensorGroup):
        print(g)
        print("")


def list_sensors(cb, parser, args):
    if args.group_name:
        group = cb.select(SensorGroup).where("name:{0}".format(args.group_name)).first()
    else:
        group = min(cb.select(SensorGroup), key=lambda x: x.id)

    print("Sensors in group {0} (id {1}):".format(group.name, group.id))
    print("  {0:40}{1:18}{2}".format("Hostname", "IP Address", "Last Checkin Time"))
    for sensor in group.sensors:
        ipaddrs = [iface.ipaddr for iface in sensor.network_interfaces if iface.ipaddr not in ("127.0.0.1", "0.0.0.0")]
        print("  {0:40}{1:18}{2}".format(sensor.hostname, ipaddrs[0], sensor.last_checkin_time))


def add_sensor_group(cb, parser, args):
    g = cb.create(SensorGroup)

    if args.site_name:
        site = cb.select(Site).where("name:{0}".format(args.site_name)).first()
        if not site:
            print("Could not find site named {0}".format(args.site_name))
            return 1
    elif args.site_id:
        site = cb.select(Site).where("id:{0}".format(args.site_id)).first()
        if not site:
            print("Could not find site ID {0}".format(args.site_id))
            return 1
    else:
        # Just pick the first site by the lowest ID
        site = min(cb.select(Site), key=lambda x: x.id)

    g.name = args.new_group_name
    g.site = site

    try:
        g.save()
    except ServerError as se:
        print("Received HTTP error {0} when attempting to add sensor group: {1}".format(se.error_code, str(se)))
    except Exception as e:
        print("Could not add sensor group: {0:s}".format(str(e)))
    else:
        log.debug("New Sensor Group: {0:s}".format(str(g)))
        print("Added sensor group. New sensor group ID is {0:d}".format(g.id))


def delete_sensor_group(cb, parser, args):
    try:
        if args.id:
            attempted_to_find = "ID of {0:d}".format(args.id)
            groups = [cb.select(SensorGroup, args.id, force_init=True)]
        else:
            attempted_to_find = "name {0:s}".format(args.groupname)
            groups = cb.select(SensorGroup).where("name:{0:s}".format(args.groupname))[::]
            if not len(groups):
                raise Exception("No sensor groups match")
    except Exception as e:
        print("Could not find sensor group with {0:s}: {1:s}".format(attempted_to_find, str(e)))
        return

    num_matching_sensor_groups = len(groups)
    if num_matching_sensor_groups > 1 and not args.force:
        print("{0:d} sensor groups match {1:s} and --force not specified. No action taken."
              .format(num_matching_sensor_groups, attempted_to_find))
        return

    for g in groups:
        try:
            g.delete()
        except Exception as e:
            print("Could not delete sensor group with {0:s}: {1:s}".format(attempted_to_find, str(e)))
        else:
            print("Deleted sensor group id {0:d} with name {1:s}".format(g.id, g.name))


def main():
    parser = build_cli_parser()
    commands = parser.add_subparsers(help="Sensor Group commands", dest="command_name")

    commands.add_parser("list", help="List all configured sensor groups")

    add_command = commands.add_parser("add", help="Add new sensor group")
    add_command.add_argument("-n", "--name", action="store", help="Sensor group name", required=True,
                             dest="new_group_name")
    site_group = add_command.add_mutually_exclusive_group(required=False)
    site_group.add_argument("-s", "--site", action="store", help="Site name", dest="site_name")
    site_group.add_argument("-i", "--site-id", action="store", help="Site ID", dest="site_id")

    del_command = commands.add_parser("delete", help="Delete sensor groups")
    del_sensor_group_specifier = del_command.add_mutually_exclusive_group(required=True)
    del_sensor_group_specifier.add_argument("-i", "--id", type=int, help="ID of sensor group to delete")
    del_sensor_group_specifier.add_argument("-n", "--name",
                                            help="Name of sensor group to delete. Specify --force to delete"
                                            " multiple sensor groups that have the same name", dest="groupname")
    del_command.add_argument("--force",
                             help="If NAME matches multiple sensor groups, delete all matching sensor groups",
                             action="store_true", default=False)

    list_sensors_command = commands.add_parser("list-sensors", help="List all sensors in a sensor group")
    list_sensors_command.add_argument("-n", "--name", action="store", help="Sensor group name", required=False,
                                      dest="group_name")

    args = parser.parse_args()
    cb = get_cb_response_object(args)

    if args.command_name == "list":
        return list_sensor_groups(cb, parser, args)
    elif args.command_name == "add":
        return add_sensor_group(cb, parser, args)
    elif args.command_name == "delete":
        return delete_sensor_group(cb, parser, args)
    elif args.command_name == "list-sensors":
        return list_sensors(cb, parser, args)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

import sys
from cbapi.example_helpers import build_cli_parser, get_cb_defense_object
import json


def main():
    parser = build_cli_parser("Listen to real-time notifications")
    parser.add_argument("-s", type=int, help="# of seconds to sleep between polls", default=30)

    args = parser.parse_args()
    cb = get_cb_defense_object(args)

    while True:
        for notification in cb.notification_listener(args.s):
            print(json.dumps(notification, indent=2))


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

from __future__ import absolute_import
import copy
import cbapi.six as six
from .errors import ApiError, MoreThanOneResultError
from cbapi.six import iteritems
from cbapi.six.moves import range
import logging


log = logging.getLogger(__name__)


class BaseQuery(object):
    def __init__(self, query=None):
        self._query = query

    def _clone(self):
        return self.__class__(self._query)

    def all(self):
        return self._perform_query()

    def first(self):
        res = self[:1]
        if not len(res):
            return None
        return res[0]

    def one(self):
        res = self[:2]

        if len(res) == 0 or len(res) > 1:
            raise MoreThanOneResultError(message="{0:d} results found for query {1:s}"
                                         .format(len(self), str(self._query)))

        return res[0]

    def _perform_query(self):
        # This has the effect of generating an empty iterator.
        # See http://stackoverflow.com/questions/13243766/python-empty-generator-function
        return
        yield

    def __len__(self):
        return 0

    def __getitem__(self, item):
        return None

    def __iter__(self):
        return self._perform_query()


class SimpleQuery(BaseQuery):
    _multiple_where_clauses_accepted = False

    def __init__(self, cls, cb, urlobject=None, returns_fulldoc=True):
        super(SimpleQuery, self).__init__()

        self._doc_class = cls
        if not urlobject:
            self._urlobject = cls.urlobject
        else:
            self._urlobject = urlobject
        self._cb = cb
        self._full_init = False
        self._results = []
        self._query = {}
        self._sort_by = None
        self._returns_full_doc = returns_fulldoc

    def _clone(self):
        nq = self.__class__(self._doc_class, self._cb)
        nq._urlobject = self._urlobject
        nq._full_init = self._full_init
        nq._results = self._results[::]
        nq._query = copy.deepcopy(self._query)
        nq._sort_by = self._sort_by

        return nq

    def _match_query(self, i):
        for k, v in iteritems(self._query):
            if isinstance(v, six.string_types):
                v = v.lower()
            target = getattr(i, k, None)
            if target is None:
                return False
            if str(target).lower() != v:
                return False
        return True

    def _sort(self, result_set):
        if self._sort_by is not None:
            return sorted(result_set, key=lambda x: getattr(x, self._sort_by, 0), reverse=True)
        else:
            return result_set

    @property
    def results(self):
        if not self._full_init:
            self._results = []
            for item in self._cb.get_object(self._urlobject, default=[]):
                t = self._doc_class.new_object(self._cb, item, full_doc=self._returns_full_doc)
                if self._match_query(t):
                    self._results.append(t)
            self._results = self._sort(self._results)
            self._full_init = True

        return self._results

    def __len__(self):
        return len(self.results)

    def __getitem__(self, item):
        if isinstance(item, slice):
            return [self.results[ii] for ii in range(*item.indices(len(self)))]
        elif isinstance(item, int):
            return self.results[item]
        else:
            raise TypeError("Invalid argument type")

    def where(self, new_query):
        if self._query and not self._multiple_where_clauses_accepted:
            raise ApiError("Cannot have multiple 'where' clauses")

        nq = self._clone()
        field, value = new_query.split(':', 1)
        nq._query[field] = value
        nq._full_init = False
        return nq

    def and_(self, new_query):
        if not self._multiple_where_clauses_accepted:
            raise ApiError("Cannot have multiple 'where' clauses")
        return self.where(new_query)

    def _perform_query(self):
        for item in self.results:
            yield item

    def sort(self, new_sort):
        nq = self._clone()
        nq._sort_by = new_sort
        return nq


class PaginatedQuery(BaseQuery):
    def __init__(self, cls, cb, query=None):
        super(PaginatedQuery, self).__init__(query)
        self._doc_class = cls
        self._cb = cb
        # TODO: this should be subject to a TTL
        self._total_results = 0
        self._count_valid = False
        self._batch_size = 100

    def _clone(self):
        nq = self.__class__(self._doc_class, self._cb, query=self._query)
        nq._batch_size = self._batch_size
        return nq

    def __len__(self):
        if self._count_valid:
            return self._total_results
        return self._count()

    def __getitem__(self, item):
        if isinstance(item, slice):
            if item.step and item.step != 1:
                raise ValueError("steps not supported")

            must_count_result_set = False
            if item.start is None or item.start == 0:
                start = 0
            else:
                start = item.start
                if item.start < 0:
                    must_count_result_set = True

            if item.stop is None or item.stop == 0:
                numrows = 0
            else:
                numrows = item.stop - start
                if item.stop < 0:
                    must_count_result_set = True
                elif numrows <= 0:
                    return []

            if must_count_result_set:
                log.debug('Must count result set')
                # general case
                item_range = range(*item.indices(len(self)))
                if not len(item_range):
                    return []

                start, numrows = item_range[0], len(item_range)

            try:
                return list(self._perform_query(start, numrows))
            except StopIteration:
                return []
        elif isinstance(item, int):
            if item < 0:
                item += len(self)
            if item < 0:
                return None

            try:
                return next(self._perform_query(item, 1))
            except StopIteration:
                return None
        else:
            raise TypeError("invalid type")

    def _perform_query(self, start=0, numrows=0):
        for item in self._search(start=start, rows=numrows):
            yield self._doc_class.new_object(self._cb, item)

    def batch_size(self, new_batch_size):
        nq = self._clone()
        nq._batch_size = new_batch_size
        return nq



"""Parses STIX observables from the XML data returned by the TAXII server.

The following IOC types are extracted from STIX data:

* MD5 Hashes
* Domain Names
* IP-Addresses
* IP-Address Ranges
"""

from cybox.objects.domain_name_object import DomainName
from cybox.objects.address_object import Address
from cybox.objects.file_object import File
from cybox.objects.uri_object import URI
from lxml import etree
from io import BytesIO
from stix.core import STIXPackage

import logging
import string
import socket
import uuid
import time
import datetime
import dateutil
import dateutil.tz

from cabby.constants import (
    CB_STIX_XML_111, CB_CAP_11, CB_SMIME,
    CB_STIX_XML_10, CB_STIX_XML_101, CB_STIX_XML_11, CB_XENC_122002)

CB_STIX_XML_12 = 'urn:stix.mitre.org:xml:1.2'

BINDING_CHOICES = [CB_STIX_XML_111, CB_CAP_11, CB_SMIME, CB_STIX_XML_12,
                   CB_STIX_XML_10, CB_STIX_XML_101, CB_STIX_XML_11,
                   CB_XENC_122002]


logger = logging.getLogger(__name__)


domain_allowed_chars = string.printable[:-6]  # Used by validate_domain_name function


def validate_domain_name(domain_name):
    """Validates a domain name to ensure validity and saneness.

    Args:
        domain_name: Domain name string to check.

    Returns:
        True if checks pass, False otherwise.
    """

    if len(domain_name) > 255:
        logger.warn(
            "Excessively long domain name {} in IOC list".format(domain_name))
        return False

    if not all([c in domain_allowed_chars for c in domain_name]):
        logger.warn("Malformed domain name {} in IOC list".format(domain_name))
        return False

    parts = domain_name.split('.')
    if not parts:
        logger.warn("Empty domain name found in IOC list")
        return False

    for part in parts:
        if len(part) < 1 or len(part) > 63:
            logger.warn("Invalid label length {} in domain name {} for report %s".format(
                part, domain_name))
            return False

    return True


def validate_md5sum(md5):
    """Validates md5sum.

    Args:
        md5sum: md5sum to check.

    Returns:
        True if checks pass, False otherwise.
    """

    if 32 != len(md5):
        logger.warn("Invalid md5 length for md5 {}".format(md5))
        return False
    if not md5.isalnum():
        logger.warn("Malformed md5 {} in IOC list".format(md5))
        return False
    for c in "ghijklmnopqrstuvwxyz":
        if c in md5 or c.upper() in md5:
            logger.warn("Malformed md5 {} in IOC list".format(md5))
            return False

    return True


def sanitize_id(id):
    """Removes unallowed chars from an ID.

    Ids may only contain a-z, A-Z, 0-9, - and must have one character.

    Args:
        id: the ID to be sanitized.

    Returns:
        A sanitized ID.
    """

    return id.replace(':', '-')


def validate_ip_address(ip_address):
    """Validates an IPv4 address."""

    try:
        socket.inet_aton(ip_address)
        return True
    except socket.error:
        return False


def cybox_parse_observable(observable, indicator, timestamp, score):
    """Parses a cybox observable and returns a list containing a report dictionary.

    cybox is a open standard language encoding info about cyber observables.

    Args:
        observable: the cybox obserable to parse.

    Returns:
        A report dictionary if the cybox observable has props of type:

            cybox.objects.address_object.Address,
            cybox.objects.file_object.File, or
            cybox.objects.domain_name_object.DomainName.

        Otherwise it will return an empty list.

    """
    reports = []

    if observable.object_ and observable.object_.properties:
        props = observable.object_.properties
        logger.debug("{0} has props type: {1}".format(indicator, type(props)))
    else:
        logger.debug("{} has no props; skipping".format(indicator))
        return reports

    #
    # sometimes the description is None
    #
    description = ''
    if observable.description and observable.description.value:
        description = str(observable.description.value)

    #
    # if description is an empty string, then use the indicator's description
    # NOTE: This was added for RecordedFuture
    #

    if not description and indicator and indicator.description and indicator.description.value:
        description = str(indicator.description.value)

    #
    # if description is still empty, use the indicator's title
    #
    if not description and indicator and indicator.title:
        description = str(indicator.title)

    #
    # use the first reference as a link
    # This was added for RecordedFuture
    #
    link = ''
    if indicator and indicator.producer and indicator.producer.references:
        for reference in indicator.producer.references:
            link = reference
            break
    else:
        if indicator and indicator.title:
            split_title = indicator.title.split()
            if split_title[2][0:8] == 'https://' or split_title[2][0:7] == 'http://':
                link = split_title[2]
        elif observable and observable.title:
            split_title = observable.title.split()
            if split_title[1][0:8] == 'https://' or split_title[1][0:7] == 'http://':
                link = split_title[1]
        else:
            link = ''


    #
    # Sometimes the title is None, so generate a random UUID
    #

    if observable and observable.title:
        title = observable.title
    else:
        title = str(uuid.uuid4())

    # ID must be unique. Collisions cause 500 error on Carbon Black backend
    id = str(uuid.uuid4())

    if type(props) == DomainName:
        # go into domainname function
        reports = parse_domain_name_observable(observable, props, id, description, title, timestamp, link, score)

    elif type(props) == Address:
        reports = parse_address_observable(observable, props, id, description, title, timestamp, link, score)

    elif type(props) == File:
        reports = parse_file_observable(observable, props, id, description, title, timestamp, link, score)

    elif type(props) == URI:
        reports = parse_uri_observable(observable, props, id, description, title, timestamp, link, score)

    else:
        return reports

    return reports

def parse_uri_observable(observable, props, id, description, title, timestamp, link, score):

    reports = []

    if props.value and props.value.value:

        iocs = {'dns': []}
        #
        # Sometimes props.value.value is a list
        #

        if type(props.value.value) is list:
            for domain_name in props.value.value:
                if validate_domain_name(domain_name.strip()):
                    iocs['dns'].append(domain_name.strip())
        else:
            domain_name = props.value.value.strip()
            if validate_domain_name(domain_name):
                iocs['dns'].append(domain_name)
                
        if len(iocs['dns']) > 0:
            reports.append({'iocs_v2': iocs,
                            'id': sanitize_id(id),
                            'description': description,
                            'title': title,
                            'timestamp': timestamp,
                            'link': link,
                            'score': score})
    return reports


def parse_domain_name_observable(observable, props, id, description, title, timestamp, link, score):

    reports = []
    if props.value and props.value.value:
        iocs = {'dns': []}
        #
        # Sometimes props.value.value is a list
        #

        if type(props.value.value) is list:
            for domain_name in props.value.value:
                if validate_domain_name(domain_name.strip()):
                    iocs['dns'].append(domain_name.strip())
        else:
            domain_name = props.value.value.strip()
            if validate_domain_name(domain_name):
                iocs['dns'].append(domain_name)

        if len(iocs['dns']) > 0:
            reports.append({'iocs_v2': iocs,
                            'id': sanitize_id(id),
                            'description': description,
                            'title': title,
                            'timestamp': timestamp,
                            'link': link,
                            'score': score})
    return reports


def parse_address_observable(observable, props, id, description, title, timestamp, link, score):

    reports = []
    if props.category == 'ipv4-addr' and props.address_value:
        iocs = {'ipv4': []}

        #
        # Sometimes props.address_value.value is a list vs a string
        #
        if type(props.address_value.value) is list:
            for ip in props.address_value.value:
                if validate_ip_address(ip.strip()):
                    iocs['ipv4'].append(ip.strip())
        else:
            ipv4 = props.address_value.value.strip()
            if validate_ip_address(ipv4):
                iocs['ipv4'].append(ipv4)

        if len(iocs['ipv4']) > 0:
            reports.append({'iocs_v2': iocs,
                            'id': sanitize_id(observable.id_),
                            'description': description,
                            'title': title,
                            'timestamp': timestamp,
                            'link': link,
                            'score': score})

    return reports


def parse_file_observable(observable, props, id, description, title, timestamp, link, score):

    reports = []
    iocs = {'md5': []}
    if props.md5:
        if type(props.md5) is list:
            for md5 in props.md5:
                if validate_md5sum(md5.strip()):
                    iocs['md5'].append(md5.strip())
        else:
            if hasattr(props.md5, 'value'):
                md5 = props.md5.value.strip()
            else:
                md5 = props.md5.strip()
            if validate_md5sum(md5):
                iocs['md5'].append(md5)

        if len(iocs['md5']) > 0:
            reports.append({'iocs_v2': iocs,
                            'id': sanitize_id(id),
                            'description': description,
                            'title': title,
                            'timestamp': timestamp,
                            'link': link,
                            'score': score})

    return reports


def get_stix_indicator_score(indicator, default_score):
    """Returns a digit representing the indicator score.

    Converts from "high", "medium", or "low" into a digit, if necessary.
    """

    if not indicator.confidence:
        return default_score

    confidence_val_str = str(indicator.confidence.value)
    if confidence_val_str.isdigit():
        score = int(indicator.confidence.to_dict().get("value", default_score))
        return score
    elif confidence_val_str.lower() == "high":
        return 7  # 75
    elif confidence_val_str.lower() == "medium":
        return 5  # 50
    elif confidence_val_str.lower() == "low":
        return 2  # 25
    else:
        return default_score


def get_stix_indicator_timestamp(indicator):
    timestamp = 0
    if indicator.timestamp:
        if indicator.timestamp.tzinfo:
            timestamp = int((indicator.timestamp -
                            datetime.datetime(1970, 1, 1).replace(
                                tzinfo=dateutil.tz.tzutc())).total_seconds())
        else:
            timestamp = int((indicator.timestamp -
                            datetime.datetime(1970, 1, 1)).total_seconds())
    return timestamp


def get_stix_package_timestamp(stix_package):
    timestamp = 0
    if not stix_package or not stix_package.timestamp:
        return timestamp
    try:
        timestamp = stix_package.timestamp
        timestamp = int(time.mktime(timestamp.timetuple()))
    except (TypeError, OverflowError, ValueError) as e:
        logger.warning("Problem parsing stix timestamp: {}".format(e))
    return timestamp


def parse_stix_indicators(stix_package, default_score):
    reports = []
    if not stix_package.indicators:
        return reports

    for indicator in stix_package.indicators:
        if not indicator or not indicator.observable:
            continue
        score = get_stix_indicator_score(indicator, default_score)
        timestamp = get_stix_indicator_timestamp(indicator)
        yield from cybox_parse_observable(
            indicator.observable, indicator, timestamp, score)


def parse_stix_observables(stix_package, default_score):
    reports = []
    if not stix_package.observables:
        return reports

    timestamp = get_stix_package_timestamp(stix_package)
    for observable in stix_package.observables:
        if not observable:
            continue
        yield from cybox_parse_observable(              # single element list
            observable, None, timestamp, default_score)


def sanitize_stix(stix_xml):
    ret_xml = b''
    try:
        xml_root = etree.fromstring(stix_xml)
        content = xml_root.find(
            './/{http://taxii.mitre.org/messages/taxii_xml_binding-1.1}Content')
        if content is not None and len(content) == 0 and len(list(content)) == 0:
            # Content has no children.
            # So lets make sure we parse the xml text for content and
            # re-add it as valid XML so we can parse
            _content = xml_root.find(
                "{http://taxii.mitre.org/messages/taxii_xml_binding-1.1}Content_Block/{http://taxii.mitre.org/messages/taxii_xml_binding-1.1}Content")
            if _content:
                new_stix_package = etree.fromstring(_content.text)
                content.append(new_stix_package)
        ret_xml = etree.tostring(xml_root)
    except etree.ParseError as e:
        logger.warning("Problem parsing stix: {}".format(e))
    return ret_xml


def parse_stix(stix_xml, default_score):
    reports = []
    try:
        stix_xml = sanitize_stix(stix_xml)
        bio = BytesIO(stix_xml)
        stix_package = STIXPackage.from_xml(bio)
        if not stix_package:
            logger.warning("Could not parse STIX xml")
            return reports
        if not stix_package.indicators and not stix_package.observables:
            logger.info("No indicators or observables found in stix_xml")
            return reports
        yield from parse_stix_indicators(stix_package, default_score)
        yield from parse_stix_observables(stix_package, default_score)
    except etree.XMLSyntaxError as e:
        logger.warning("Problem parsing stix: {}".format(e))
        return reports



#!/usr/bin/env python

import sys
from cbapi.response.models import Alert
from cbapi.example_helpers import build_cli_parser, get_cb_response_object
import time


def main():
    parser = build_cli_parser("Bulk resolve alerts")
    parser.add_argument("--query", action="store", default="", required=True,
                        help="The query string of alerts to resolve. All matching alerts will be resolved.")

    args = parser.parse_args()

    cb = get_cb_response_object(args)

    alert_query = cb.select(Alert).where("-status:Resolved")
    alert_query = alert_query.where(args.query)

    alert_count = len(alert_query)

    if alert_count > 0:
        print("Resolving {0:d} alerts...".format(len(alert_query)))

        alert_query.change_status("Resolved")

        print("Waiting for alert changes to take effect...")
        time.sleep(25)
        print("Complete. Resolved {0:d} alerts.".format(alert_count))
    else:
        print("Congratulations! You have no unresolved alerts!")


if __name__ == "__main__":
    sys.exit(main())



"""Connects to TAXII servers via cabby and formats the data received for dispatching to a Carbon Black feed."""

import argparse
import logging
from threatintel import ThreatIntel
from cabby.exceptions import NoURIProvidedError, ClientException
from cabby import create_client
from dataclasses import dataclass
import yaml
import os
from stix_parse import parse_stix, BINDING_CHOICES
from feed_helper import FeedHelper
from datetime import datetime
from results import AnalysisResult
import urllib3

logging.basicConfig(filename='stix.log', filemode='w', level=logging.DEBUG)

handled_exceptions = (NoURIProvidedError, ClientException)


def load_config_from_file():
    """Loads YAML formatted configuration from config.yml in working directory."""

    logging.debug("loading config from file")
    config_filename = os.path.join(os.path.dirname((os.path.abspath(__file__))), "config.yml")
    with open(config_filename, "r") as config_file:
        config_data = yaml.load(config_file, Loader=yaml.SafeLoader)
        logging.info(f"loaded config data: {config_data}")
        return config_data


@dataclass(eq=True, frozen=True)
class TaxiiSiteConfig:
    """Contains information needed to interface with a TAXII server.

    These values are loaded in from config.yml for each entry in the configuration file.
    Each TaxiiSiteConnector has its own TaxiiSiteConfig.
    """

    feed_id: str = ''
    site: str = ''
    discovery_path: str = ''
    collection_management_path: str = ''
    poll_path: str = ''
    use_https: bool = False
    ssl_verify: bool = False
    cert_file: str = None
    key_file: str = None
    default_score: int = 5  # [1,10]
    username: str = None
    password: str = None
    collections: str = '*'
    start_date: str = None
    size_of_request_in_minutes: int = 1440
    ca_cert: str = None
    http_proxy_url: str = None
    https_proxy_url: str = None
    reports_limit: int = None
    fail_limit: int = 10   # num attempts per collection for polling & parsing


class TaxiiSiteConnector():
    """Connects to and pulls data from a TAXII server."""

    def __init__(self, site_conf):
        self.config = TaxiiSiteConfig(**site_conf)
        self.client = None

    def create_taxii_client(self):
        """Connects to a TAXII server using cabby and configuration entries."""

        conf = self.config
        if not conf.start_date:
            logging.error(f"A start_date is required for site {conf.site}. Exiting.")
            return
        if not conf.ssl_verify:
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        try:
            client = create_client(conf.site,
                                   use_https=conf.use_https,
                                   discovery_path=conf.discovery_path)
            client.set_auth(username=conf.username,
                            password=conf.password,
                            verify_ssl=conf.ssl_verify,
                            ca_cert=conf.ca_cert,
                            cert_file=conf.cert_file,
                            key_file=conf.key_file)

            proxy_dict = dict()
            if conf.http_proxy_url:
                proxy_dict['http'] = conf.http_proxy_url
            if conf.https_proxy_url:
                proxy_dict['https'] = conf.https_proxy_url
            if proxy_dict:
                client.set_proxies(proxy_dict)

            self.client = client

        except handled_exceptions as e:
            logging.error(f"Error creating client: {e}")

    def create_uri(self, config_path):
        """Formats a URI for discovery, collection, or polling of a TAXII server.

        Args:
            config_path: A URI path to a TAXII server's discovery, collection, or polling service. Defined in config.yml configuration file.

        Returns:
            A full URI to one of a TAXII server's service paths.
        """

        uri = None
        if self.config.site and config_path:
            if self.config.use_https:
                uri = 'https://'
            else:
                uri = 'http://'
            uri = uri + self.config.site + config_path
        return uri

    def query_collections(self):
        """Returns a list of STIX collections available to the user to poll."""

        collections = []
        try:
            uri = self.create_uri(self.config.collection_management_path)
            collections = self.client.get_collections(
                uri=uri)  # autodetect if uri=None
            for collection in collections:
                logging.info(f"Collection: {collection.name}, {collection.type}")
        except handled_exceptions as e:
            logging.warning(f"Problem fetching collections from taxii server: {e}")
        return collections

    def poll_server(self, collection, feed_helper):
        """Returns a STIX content block for a specific TAXII collection.

        Args:
            collection: Name of a TAXII collection to poll.
            feed_helper: FeedHelper object.
        """

        content_blocks = []
        uri = self.create_uri(self.config.poll_path)
        try:
            logging.info(f"Polling Collection: {collection.name}")
            content_blocks = self.client.poll(
                uri=uri,
                collection_name=collection.name,
                begin_date=feed_helper.start_date,
                end_date=feed_helper.end_date,
                content_bindings=BINDING_CHOICES)
        except handled_exceptions as e:
            logging.warning(f"problem polling taxii server: {e}")
        return content_blocks

    def parse_collection_content(self, content_blocks):
        """Yields a formatted report dictionary for each STIX content_block.

        Args:
            content_block: A chunk of STIX data from the TAXII collection being polled.
        """

        for block in content_blocks:
            yield from parse_stix(block.content, self.config.default_score)

    def import_collection(self, collection):
        """Polls a single TAXII server collection.

        Starting at the start_date set in config.yml, a FeedHelper object will continue to grab chunks
        of data from a collection until the report limit is reached or we reach the current datetime.

        Args:
            collection: Name of a TAXII collection to poll.

        Yields:
            Formatted report dictionaries from parse_collection_content(content_blocks)
            for each content_block pulled from a single TAXII collection.
        """

        num_times_empty_content_blocks = 0
        advance = True
        reports_limit = self.config.reports_limit
        logging.info(f"reports limit: {reports_limit}")
        if not self.config.size_of_request_in_minutes:
            size_of_request_in_minutes = 1440
        else:
            size_of_request_in_minutes = self.config.size_of_request_in_minutes
        feed_helper = FeedHelper(self.config.start_date,
                                 size_of_request_in_minutes)
        # config parameters `start_date` and `size_of_request_in_minutes` tell this Feed Helper
        # where to start polling in the collection, and then will advance polling in chunks of
        # `size_of_request_in_minutes` until we hit the most current `content_block`,
        # or reports_limit is reached.
        while feed_helper.advance():
            num_reports = 0
            num_times_empty_content_blocks = 0
            content_blocks = self.poll_server(collection, feed_helper)
            reports = self.parse_collection_content(content_blocks)
            for report in reports:
                yield report
                num_reports += 1
                if reports_limit is not None and num_reports >= reports_limit:
                    logging.info(f"Reports limit of {self.config.reports_limit} reached")
                    advance = False
                    break

            if not advance:
                break
            if collection.type == 'DATA_SET':  # data is unordered, not a feed
                logging.info(f"collection:{collection}; type data_set; breaking")
                break
            if num_reports == 0:
                num_times_empty_content_blocks += 1
            if num_times_empty_content_blocks > self.config.fail_limit:
                logging.error('Max fail limit reached; Exiting.')
                break
            if reports_limit is not None:
                reports_limit -= num_reports

    def import_collections(self, available_collections):
        """Polls each desired collection specified in config.yml.

        Args:
            available_collections: list of collections available to a TAXII server user.

        Yields:
            From import_collection(self, collection) for each desired collection.
        """

        if not self.config.collections:
            desired_collections = '*'
        else:
            desired_collections = self.config.collections

        desired_collections = [x.strip()
                               for x in desired_collections.lower().split(',')]

        want_all = True if '*' in desired_collections else False

        for collection in available_collections:
            if collection.type != 'DATA_FEED' and collection.type != 'DATA_SET':
                logging.debug(f"collection:{collection}; type not feed or data")
                continue
            if not collection.available:
                logging.debug(f"collection:{collection} not available")
                continue
            if want_all or collection.name.lower() in desired_collections:
                yield from self.import_collection(collection)

    def generate_reports(self):
        """Returns a list of report dictionaries for each desired collection specified in config.yml."""

        reports = []

        self.create_taxii_client()
        if not self.client:
            logging.error('Unable to create taxii client.')
            return reports

        available_collections = self.query_collections()
        if not available_collections:
            logging.warning('Unable to find any collections.')
            return reports

        reports = self.import_collections(available_collections)
        if not reports:
            logging.warning('Unable to import collections.')
            return reports

        return reports


class StixTaxii():
    """Allows for interfacing with multiple TAXII servers.

    Instantiates separate TaxiiSiteConnector objects for each site specified in config.yml.
    Formats report dictionaries into AnalysisResult objects with formatted IOC_v2 attirbutes.
    Sends AnalysisResult objects to ThreatIntel.push_to_cb for dispatching to a feed.
    """

    def __init__(self, site_confs):
        self.config = site_confs
        self.client = None

    def result(self, **kwargs):
        """Returns a new AnalysisResult with the given fields populated."""

        result = AnalysisResult(**kwargs).normalize()
        return result

    def configure_sites(self):
        """Creates a TaxiiSiteConnector for each site in config.yml"""

        self.sites = {}
        try:
            for site_name, site_conf in self.config['sites'].items():
                self.sites[site_name] = TaxiiSiteConnector(site_conf)
                logging.info(f"loaded site {site_name}")
        except handled_exceptions as e:

            logging.error(f"Error in parsing config file: {e}")

    def format_report(self, reports):
        """Converts a dictionary into an AnalysisResult.

        Args:
            reports: list of report dictionaries containing an id, title, description, timestamp, score, link, and iocs_v2.

        Yields:
            An AnalysisResult for each report dictionary.
        """

        for report in reports:
            try:
                analysis_name = report['id']
                title = report['title']
                description = report['description']
                scan_time = datetime.fromtimestamp(report['timestamp'])
                score = report['score']
                link = report['link']
                ioc_dict = report['iocs_v2']
                result = self.result(
                                     analysis_name=analysis_name,
                                     scan_time=scan_time,
                                     score=score,
                                     title=title,
                                     description=description)
                for ioc_key, ioc_val in ioc_dict.items():
                    result.attach_ioc_v2(values=ioc_val, field=ioc_key, link=link)
            except handled_exceptions as e:
                logging.warning(f"Problem in report formatting: {e}")
                result = self.result(
                    analysis_name="exception_format_report", error=True)
            yield result

    def collect_and_send_reports(self):
        """Collects and sends formatted reports to ThreatIntel.push_to_cb for validation and dispatching to a feed."""

        self.configure_sites()
        ti = ThreatIntel()
        for site_name, site_conn in self.sites.items():
            logging.info(f"Talking to {site_name} server")
            reports = site_conn.generate_reports()
            if not reports:
                logging.error(f"No reports generated for {site_name}")
                continue
            else:
                try:
                    ti.push_to_cb(feed_id=site_conn.config.feed_id, results=self.format_report(reports))
                except Exception as e:
                    logging.error(f"Failed to push reports to feed {site_conn.config.feed_id}: {e}")


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='Modify configuration values via command line.')
    parser.add_argument('--site_start_date', metavar='s', nargs='+',
                        help='the site name and desired start date to begin polling from')
    args = parser.parse_args()

    config = load_config_from_file()

    if args.site_start_date:
        for index in range(len(args.site_start_date)):
            arg = args.site_start_date[index]
            if arg in config['sites']:  # if we see a name that matches a site Name
                try:
                    new_time = datetime.strptime(args.site_start_date[index+1], "%Y-%m-%d %H:%M:%S")
                    config['sites'][arg]['start_date'] = new_time
                    logging.info(f"Updated the start_date for {arg} to {new_time}")
                except ValueError as e:
                    logging.error(f"Failed to update start_date for {arg}: {e}")
    stix_taxii = StixTaxii(config)
    stix_taxii.collect_and_send_reports()



#!/usr/bin/env python
#

import sys
from cbapi.response.models import Sensor
from cbapi.example_helpers import build_cli_parser, get_cb_response_object
import logging
import csv
import traceback

log = logging.getLogger(__name__)


def export_sensors(cb, export_file_name, export_fields, query):
    print("Starting CbR Sensor Export")
    if query:
        sensors = list(cb.select(Sensor).where(query).all())
    else:
        sensors = list(cb.select(Sensor))
    with open(export_file_name, "w", encoding="utf8") as csv_file:
        csv_writer = csv.writer(csv_file, delimiter=',', lineterminator='\n')
        csv_writer.writerow(export_fields)
        for sensor in sensors:
            try:
                row = [getattr(sensor, field) for field in export_fields]
                csv_writer.writerow(row)
            except Exception as e:
                print("Exception {1} caused sensor export to fail for {0}".format(sensor.hostname, str(e)))
                traceback.format_exc(0)
    print("Export finished, exported {0} sensors to {1}".format(len(sensors), export_file_name))


def main():
    parser = build_cli_parser(description="Export CbR Sensors from your environment as CSV")
    parser.add_argument("--output", "-o", dest="exportfile", help="The file to export to", required=True)
    parser.add_argument("--fields", "-f", dest="exportfields", help="The fields to export",
                        default="id,hostname,group_id,network_interfaces,os_environment_display_string,"
                        "build_version_string,network_isolation_enabled,last_checkin_time",
                        required=False)
    parser.add_argument("--query", "-q", dest="query", help="optional query to filter exported sensors", required=False)
    args = parser.parse_args()
    cb = get_cb_response_object(args)
    export_fields = args.exportfields.split(",")
    return export_sensors(cb, export_file_name=args.exportfile, export_fields=export_fields, query=args.query)


if __name__ == "__main__":
    sys.exit(main())



"""
cbapi-python
"""

from setuptools import setup
import sys

packages = [
    'cbapi',
    'cbapi.protection',
    'cbapi.response',
    'cbapi.cache',
    'cbapi.psc',
    'cbapi.psc.defense',
    'cbapi.psc.threathunter',
    'cbapi.psc.livequery'
]

install_requires = [
    'requests',
    'attrdict',
    'cachetools',
    'pyyaml',
    'pika',
    'prompt_toolkit',
    'pygments',
    'python-dateutil',
    'protobuf',
    'solrq',
    'validators'
]

tests_requires = [
    'pytest<=5.0'
]

if sys.version_info < (2, 7):
    install_requires.extend(['simplejson', 'total-ordering', 'ordereddict'])
if sys.version_info < (3, 0):
    install_requires.extend(['futures'])

setup(
    name='cbapi',
    version='1.6.2',
    url='https://github.com/carbonblack/cbapi-python',
    license='MIT',
    author='Carbon Black',
    author_email='dev-support@carbonblack.com',
    description='Carbon Black REST API Python Bindings',
    packages=packages,
    include_package_data=True,
    package_dir={'': 'src'},
    zip_safe=False,
    platforms='any',
    install_requires=install_requires,
    tests_requires=tests_requires,
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    scripts=['bin/cbapi-response', 'bin/cbapi-protection', 'bin/cbapi-defense', 'bin/cbapi', 'bin/cbapi-psc']
)



#!/usr/bin/env python
#
# The MIT License (MIT)
##
# Copyright (c) 2015 Bit9 + Carbon Black
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# -----------------------------------------------------------------------------
#  <Short Description>
#
# Usage: process_netconn_rate.py [options]
#
# High avg. netconn/second alert
#
# Options:
#  -h, --help            show this help message and exit
#  -g GT_COUNT, --gt-count=GT_COUNT
#                        Filter processes with greater than [--gt-count]
#                        network events
#  -r CONN_RATE, --rate=CONN_RATE
#                        Alert on processes with more than [--rate] network
#                        connections per second
#  -s, --skip_unknown    Skip processes with unknown start or last update
#
# simple script to detect processes with a high rate of network connections per second
#
#   -Optionally control the minimum total number of connections with the '-g' flag
#       (defaults to > 100)
#   -Optionally control the rate of connections that will generate an alert with the '-r'
#       flag (defaults to > 100 connections/second)
#   -In testing there were some cases that could not identify either a process start time
#       or a process last update time the '-s' flag provide the option to show or hide
#       such processes (defaults to showing those processes)
#
#

import sys
from cbapi.response.models import Process
from cbapi.example_helpers import build_cli_parser, get_cb_response_object


def main():
    parser = build_cli_parser("High avg. netconn/second alert")
    parser.add_argument("--skip-unknown", "-s", action="store_true", default=False, dest="skip_unknown",
                        help="Skip processes with unknown start or last update")
    parser.add_argument("--rate", "-r", type=float, default=100.0, dest="conn_rate",
                        help="Alert on processes with more than [--rate] network connections per second")
    parser.add_argument("--gt-count", "-g", type=int, default=100, dest="gt_count",
                        help="Filter processes with greater than [--gt-count] network events")

    args = parser.parse_args()
    cb = get_cb_response_object(args)

    for proc in cb.select(Process).where("netconn_count:[{0:d} TO *]".format(args.gt_count)).sort("last_update desc"):
        try:
            runtime = (proc.last_update - proc.start).total_seconds()
        except Exception:
            if not args.skip_unknown:
                runtime = 1.0
            else:
                continue

        if not runtime and proc.netconn_count > 0:
            # simulate "infinity" so as to avoid a DivideByZero exception
            rate = 1000000
        else:
            rate = proc.netconn_count / float(runtime)

        if rate > args.conn_rate:
            print("{0:s}|{1:s}|{2:s}|{3:.4f}".format(proc.hostname, proc.username, proc.process_name, rate))


if __name__ == "__main__":
    sys.exit(main())



from cbapi.psc.livequery.models import Run, RunHistory
from cbapi.psc.rest_api import CbPSCBaseAPI
from cbapi.errors import CredentialError, ApiError
import logging

log = logging.getLogger(__name__)


class CbLiveQueryAPI(CbPSCBaseAPI):
    """The main entry point into the Carbon Black Cloud LiveQuery API.

    :param str profile: (optional) Use the credentials in the named profile when connecting to the Carbon Black server.
        Uses the profile named 'default' when not specified.

    Usage::

    >>> from cbapi.psc.livequery import CbLiveQueryAPI
    >>> cb = CbLiveQueryAPI(profile="production")
    """
    def __init__(self, *args, **kwargs):
        super(CbLiveQueryAPI, self).__init__(*args, **kwargs)

        if not self.credentials.get("org_key", None):
            raise CredentialError("No organization key specified")

    def _perform_query(self, cls, **kwargs):
        if hasattr(cls, "_query_implementation"):
            return cls._query_implementation(self)
        else:
            raise ApiError("All LiveQuery models should provide _query_implementation")

    def query(self, sql):
        return self.select(Run).where(sql=sql)

    def query_history(self, query=None):
        return self.select(RunHistory).where(query)



#!/usr/bin/env python

import sys

from cbapi.example_helpers import build_cli_parser, get_cb_threathunter_object
from cbapi.psc.threathunter import Event
import json
import csv


def main():
    parser = build_cli_parser("Query processes")
    parser.add_argument("-p", type=str, help="process guid", default=None)
    parser.add_argument("-s", type=bool, help="silent mode", default=False)
    parser.add_argument("-n", type=int, help="only output N events", default=None)
    parser.add_argument("-f", type=str, help="output file name", default=None)
    parser.add_argument("-of", type=str, help="output file format: csv or json", default="json")

    args = parser.parse_args()
    cb = get_cb_threathunter_object(args)

    if not args.p:
        print("Error: Missing Process GUID to search for events with")
        sys.exit(1)

    events = cb.select(Event).where(process_guid=args.p)

    if args.n:
        events = events[0:args.n]

    if not args.s:
        for event in events:
            print("Event type: {}".format(event.event_type))
            print("\tEvent GUID: {}".format(event.event_guid))
            print("\tEvent Timestamp: {}".format(event.event_timestamp))

    if args.f is not None:
        if args.of == "json":
            with open(args.f, 'w') as outfile:
                for event in events:
                    json.dump(event.original_document, outfile)
        else:
            with open(args.f, 'w') as outfile:
                csvwriter = csv.writer(outfile)
                for event in events:
                    csvwriter.writerows(event)


if __name__ == "__main__":
    sys.exit(main())



# Exported public API for the Cb PSC API

from __future__ import absolute_import

from .rest_api import CbPSCBaseAPI
from .models import Device, Workflow, BaseAlert, WatchlistAlert, CBAnalyticsAlert, VMwareAlert, WorkflowStatus



#!/usr/bin/env python

import sys

from cbapi.example_helpers import build_cli_parser, get_cb_threathunter_object
from cbapi.psc.threathunter import Process


def main():
    parser = build_cli_parser("Query processes")
    parser.add_argument("-p", type=str, help="process guid", default=None)

    args = parser.parse_args()
    cb = get_cb_threathunter_object(args)

    if not args.p:
        print("Error: Missing Process GUID to query the process tree with")
        sys.exit(1)

    tree = cb.select(Process).where(process_guid=args.p)[0].tree()
    for idx, child in enumerate(tree.children):
        print("Child #{}".format(idx))
        print("\tName: {}".format(child.process_name))
        print("\tGUID: {}".format(child.process_guid))
        print("\tNumber of children: {}".format(len(child.children)))


if __name__ == "__main__":
    sys.exit(main())



from cbapi.utils import convert_query_params
from cbapi.query import PaginatedQuery

from cbapi.psc.rest_api import CbPSCBaseAPI
import logging
import time

log = logging.getLogger(__name__)


def convert_to_kv_pairs(q):
    k, v = q.split(':', 1)
    return k, v


class CbDefenseAPI(CbPSCBaseAPI):
    """The main entry point into the Cb Defense API.

    :param str profile: (optional) Use the credentials in the named profile when connecting to the Carbon Black server.
        Uses the profile named 'default' when not specified.

    Usage::

    >>> from cbapi import CbDefenseAPI
    >>> cb = CbDefenseAPI(profile="production")
    """
    def __init__(self, *args, **kwargs):
        super(CbDefenseAPI, self).__init__(*args, **kwargs)

    def _perform_query(self, cls, query_string=None):
        return Query(cls, self, query_string)

    def notification_listener(self, interval=60):
        """Generator to continually poll the Cb Defense server for notifications (alerts). Note that this can only
        be used with a 'SIEM' key generated in the Cb Defense console.
        """
        while True:
            for notification in self.get_notifications():
                yield notification
            time.sleep(interval)

    def get_notifications(self):
        """Retrieve queued notifications (alerts) from the Cb Defense server. Note that this can only be used
        with a 'SIEM' key generated in the Cb Defense console.

        :returns: list of dictionary objects representing the notifications, or an empty list if none available.
        """
        res = self.get_object("/integrationServices/v3/notification")
        return res.get("notifications", [])


class Query(PaginatedQuery):
    """Represents a prepared query to the Cb Defense server.

    This object is returned as part of a :py:meth:`CbDefenseAPI.select`
    operation on models requested from the Cb Defense server. You should not have to create this class yourself.

    The query is not executed on the server until it's accessed, either as an iterator (where it will generate values
    on demand as they're requested) or as a list (where it will retrieve the entire result set and save to a list).
    You can also call the Python built-in ``len()`` on this object to retrieve the total number of items matching
    the query.

    Examples::

    >>> from cbapi.psc.defense import CbDefenseAPI
    >>> cb = CbDefenseAPI()

    Notes:
        - The slicing operator only supports start and end parameters, but not step. ``[1:-1]`` is legal, but
          ``[1:2:-1]`` is not.
        - You can chain where clauses together to create AND queries; only objects that match all ``where`` clauses
          will be returned.
    """
    def __init__(self, doc_class, cb, query=None):
        super(Query, self).__init__(doc_class, cb, None)
        if query:
            self._query = [query]
        else:
            self._query = []

        self._sort_by = None
        self._group_by = None
        self._batch_size = 100

    def _clone(self):
        nq = self.__class__(self._doc_class, self._cb)
        nq._query = self._query[::]
        nq._sort_by = self._sort_by
        nq._group_by = self._group_by
        nq._batch_size = self._batch_size
        return nq

    def where(self, q):
        """Add a filter to this query.

        :param str q: Query string
        :return: Query object
        :rtype: :py:class:`Query`
        """
        nq = self._clone()
        nq._query.append(q)
        return nq

    def and_(self, q):
        """Add a filter to this query. Equivalent to calling :py:meth:`where` on this object.

        :param str q: Query string
        :return: Query object
        :rtype: :py:class:`Query`
        """
        return self.where(q)

    def prepare_query(self, args):
        if self._query:
            for qe in self._query:
                k, v = convert_to_kv_pairs(qe)
                args[k] = v

        return args

    def _count(self):
        args = {'limit': 0}
        args = self.prepare_query(args)

        query_args = convert_query_params(args)
        self._total_results = int(self._cb.get_object(self._doc_class.urlobject, query_parameters=query_args)
                                  .get("totalResults", 0))
        self._count_valid = True
        return self._total_results

    def _search(self, start=0, rows=0):
        # iterate over total result set, 1000 at a time
        args = {}
        if start != 0:
            args['start'] = start
        args['rows'] = self._batch_size

        current = start
        numrows = 0

        args = self.prepare_query(args)
        still_querying = True

        while still_querying:
            query_args = convert_query_params(args)
            result = self._cb.get_object(self._doc_class.urlobject, query_parameters=query_args)

            self._total_results = result.get("totalResults", 0)
            self._count_valid = True

            results = result.get('results', [])

            if results is None:
                log.debug("Results are None")
                if current >= 100000:
                    log.info("Max result size exceeded. Truncated to 100k.")
                break

            for item in results:
                yield item
                current += 1
                numrows += 1
                if rows and numrows == rows:
                    still_querying = False
                    break

            args['start'] = current + 1     # as of 6/2017, the indexing on the Cb Defense backend is still 1-based

            if current >= self._total_results:
                break

            if not results:
                log.debug("server reported total_results overestimated the number of results for this query by {0}"
                          .format(self._total_results - current))
                log.debug("resetting total_results for this query to {0}".format(current))
                self._total_results = current
                break



#!/usr/bin/env python

import sys

from cbapi.example_helpers import build_cli_parser, get_cb_threathunter_feed_object


def main():
    parser = build_cli_parser("Modify a CbTH feed")
    parser.add_argument("--id", type=str, help="Feed ID", default=None)
    parser.add_argument("--name", type=str, help="Feed name", default=None)
    parser.add_argument("--owner", type=str, help="Feed owner", default=None)
    parser.add_argument("--url", type=str, help="Feed provider url", default="https://example.com")
    parser.add_argument("--summary", type=str, help="Feed summary", default=None)
    parser.add_argument("--category", type=str, help="Feed category", default="Partner")
    parser.add_argument("--access", type=str, help="Feed access scope", default="private")

    args = parser.parse_args()
    cb = get_cb_threathunter_feed_object(args)

    feed = cb.feed(args.id)

    print("Before modification:")
    print("=" * 80)
    print(feed)
    print("=" * 80)

    metadata = {}
    if args.name:
        metadata["name"] = args.name
    if args.owner:
        metadata["owner"] = args.owner
    if args.url:
        metadata["provider_url"] = args.url
    if args.summary:
        metadata["summary"] = args.summary
    if args.category:
        metadata["category"] = args.category
    if args.access:
        metadata["access"] = args.access

    feed.feedinfo.update(**metadata)

    print("After modification:")
    print("=" * 80)
    print(feed)
    print("=" * 80)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

from ..oldmodels import BaseModel, immutable, MutableModel
from ..models import MutableBaseModel, CreatableModelMixin, NewBaseModel
from contextlib import closing
from distutils.version import LooseVersion

from zipfile import ZipFile
import cbapi.six as six
if six.PY3:
    from io import BytesIO as StringIO
else:
    from cStringIO import StringIO


class EnforcementLevel:
    LevelHigh = 20
    LevelMedium = 30
    LevelLow = 40
    LevelNone = 80


class ApprovalRequest(MutableModel):
    urlobject = "/api/bit9platform/v1/approvalRequest"

    ResolutionNotResolved = 0
    ResolutionRejected = 1
    ResolutionApproved = 2
    ResolutionRuleChange = 3
    ResolutionInstaller = 4
    ResolutionUpdater = 5
    ResolutionPublisher = 6
    ResolutionOther = 7

    StatusSubmitted = 1
    StatusOpen = 2
    StatusClosed = 3

    def __init__(self, cb, model_unique_id, initial_data=None):
        super(ApprovalRequest, self).__init__(cb, model_unique_id, initial_data)

    @property
    def fileCatalog(self):
        return self._join(FileCatalog, "fileCatalogId")

    @property
    def installerFileCatalog(self):
        return self._join(FileCatalog, "installerFileCatalogId")

    @property
    def processFileCatalog(self):
        return self._join(FileCatalog, "processFileCatalogId")

    @property
    def computer(self):
        return self._join(Computer, "computerId")


class Certificate(MutableModel):
    urlobject = "/api/bit9platform/v1/certificate"

    StateUnapproved = 1
    StateApproved = 2
    StateBanned = 3
    StateMixed = 4

    def __init__(self, cb, model_unique_id, initial_data=None):
        super(Certificate, self).__init__(cb, model_unique_id, initial_data)

    @property
    def parent(self):
        return self._join(Certificate, "parentCertificateId")

    @property
    def publisher(self):
        return self._join(Publisher, "publisherId")

    @property
    def firstSeenComputer(self):
        return self._join(Computer, "firstSeenComputerId")


class Computer(MutableBaseModel):
    urlobject = "/api/bit9platform/v1/computer"

    def __init__(self, cb, model_unique_id, initial_data=None):
        super(Computer, self).__init__(cb, model_unique_id, initial_data)

    def _build_api_request_uri(self, http_method="GET"):
        base_uri = super(Computer, self)._build_api_request_uri(http_method=http_method)
        args = []

        if http_method == "PUT":
            if any(n in self._dirty_attributes
                   for n in ["debugLevel", "kernelDebugLevel", "debugFlags", "debugDuration", "ccLevel", "ccFlags"]):
                args.append("changeDiagnostics=true")
            if any(n in self._dirty_attributes
                   for n in ["template", "templateCloneCleanupMode", "templateCloneCleanupTime",
                             "templateCloneCleanupTimeScale", "templateTrackModsOnly"]):
                args.append("changeTemplate=true")
            if "tamperProtectionActive" in self._dirty_attributes:
                if self.get("tamperProtectionActive", True):
                    args.append("newTamperProtectionActive=true")
                else:
                    args.append("newTamperProtectionActive=false")

            if args:
                base_uri += "?{0}".format("&".join(args))

        return base_uri

    @property
    def policy(self):
        return self._join(Policy, "policyId")

    @policy.setter
    def policy(self, new_policy_id):
        self.policyId = new_policy_id

    @property
    def fileInstances(self):
        return self._cb.select(FileInstance).where("computerId:{0:d}".format(self.id))

    @property
    def templateComputer(self):
        return self._join(Computer, "templateComputerId")

    def resetCLIPassword(self):
        url = self._build_api_request_uri() + "?resetCLIPassword=true"
        self._cb.put_object(url, {})
        self.refresh()
        return getattr(self, "CLIPassword")


class Connector(MutableBaseModel, CreatableModelMixin):
    urlobject = "/api/bit9platform/v1/connector"
    swagger_meta_file = "protection/models/connector.yaml"

    @property
    def pendingAnalyses(self):
        return self._cb.select(PendingAnalysis).where("connectorId:{0:d}".format(self.id))


class DriftReport(NewBaseModel):
    urlobject = "/api/bit9platform/v1/driftReport"

    @classmethod
    def _minimum_server_version(cls):
        return LooseVersion("8.0")


class DriftReportContents(NewBaseModel):
    urlobject = "/api/bit9platform/v1/driftReportContents"

    @classmethod
    def _minimum_server_version(cls):
        return LooseVersion("8.0")


class Event(NewBaseModel):
    urlobject = "/api/bit9platform/v1/event"

    def __init__(self, cb, model_unique_id, initial_data=None):
        super(Event, self).__init__(cb, model_unique_id, initial_data)

    @property
    def fileCatalog(self):
        return self._join(FileCatalog, "fileCatalogId")


class FileAnalysis(MutableModel):
    urlobject = "/api/bit9platform/v1/fileAnalysis"

    def __init__(self, cb, model_unique_id, initial_data=None):
        super(FileAnalysis, self).__init__(cb, model_unique_id, initial_data)


class FileCatalog(MutableBaseModel):
    urlobject = "/api/bit9platform/v1/fileCatalog"

    def __init__(self, cb, model_unique_id, initial_data=None):
        super(FileCatalog, self).__init__(cb, model_unique_id, initial_data)

    @property
    def computer(self):
        return self._cb.select(Computer, self.computerId)

    @property
    def publisher(self):
        return self._cb.select(Publisher, self.publisherId)

    @property
    def certificate(self):
        return self._cb.select(Certificate, self.certificateId)

    @property
    def fileHash(self):
        return getattr(self, "md5", None) or getattr(self, "sha1", None) or getattr(self, "sha256", None)


class FileInstance(MutableBaseModel):
    urlobject = "/api/bit9platform/v1/fileInstance"
    _change_object_http_method = "POST"

    def __init__(self, cb, model_unique_id, initial_data=None):
        super(FileInstance, self).__init__(cb, model_unique_id, initial_data)

    @property
    def computer(self):
        return self._join(Computer, "computerId")

    @property
    def fileCatalog(self):
        return self._join(FileCatalog, "fileCatalogId")


@immutable
class FileInstanceDeleted(BaseModel):
    urlobject = "/api/bit9platform/v1/fileInstanceDeleted"

    def __init__(self, cb, model_unique_id, initial_data=None):
        super(FileInstanceDeleted, self).__init__(cb, model_unique_id, initial_data)


@immutable
class FileInstanceGroup(BaseModel):
    urlobject = "/api/bit9platform/v1/fileInstanceGroup"

    def __init__(self, cb, model_unique_id, initial_data=None):
        super(FileInstanceGroup, self).__init__(cb, model_unique_id, initial_data)


class FileRule(MutableBaseModel, CreatableModelMixin):
    urlobject = "/api/bit9platform/v1/fileRule"
    swagger_meta_file = "protection/models/fileRule.yaml"

    StateUnapproved = 1
    StateApproved = 2
    StateBanned = 3

    SourceTypeManual = 1
    SourceTypeTrustedDirectory = 2
    SourceTypeReputation = 3
    SourceTypeImported = 4
    SourceTypeExternal = 5
    SourceTypeEventRule = 6
    SourceTypeApplicationTemplate = 7
    SourceTypeUnifiedManagement = 8

    PlatformWindows = 1
    PlatformMac = 2
    PlatformLinux = 4

    @property
    def fileCatalog(self):
        return self._join(FileCatalog, "fileCatalogId")

    @property
    def createdByUser(self):
        return self._join(User, "createdByUserId")


class FileUpload(MutableModel):
    urlobject = "/api/bit9platform/v1/fileUpload"

    def __init__(self, cb, model_unique_id, initial_data=None):
        super(FileUpload, self).__init__(cb, model_unique_id, initial_data)

    @property
    def file(self):
        with closing(self._cb.session.get(self._build_api_request_uri() + "?downloadFile=true", stream=True)) as r:
            z = StringIO(r.content)
            zf = ZipFile(z)
            fp = zf.open(zf.filelist[0], "r")
            return fp


class GrantedUserPolicyPermission(NewBaseModel):
    urlobject = "/api/bit9platform/v1/grantedUserPolicyPermission"

    @classmethod
    def _minimum_server_version(cls):
        return LooseVersion("8.0")


@immutable
class InternalEvent(BaseModel):
    urlobject = "/api/bit9platform/v1/internalEvent"

    def __init__(self, cb, model_unique_id, initial_data=None):
        super(InternalEvent, self).__init__(cb, model_unique_id, initial_data)


@immutable
class MeteredExecution(BaseModel):
    urlobject = "/api/bit9platform/v1/meteredExecution"

    def __init__(self, cb, model_unique_id, initial_data=None):
        super(MeteredExecution, self).__init__(cb, model_unique_id, initial_data)


class Notification(MutableBaseModel, CreatableModelMixin):
    ResultNotAvailable = 0
    ResultClean = 1
    ResultPotentialThreat = 2
    ResultMalicious = 3

    urlobject = "/api/bit9platform/v1/notification"
    swagger_meta_file = "protection/models/notification.yaml"


@immutable
class Notifier(BaseModel):
    urlobject = "/api/bit9platform/v1/notifier"

    def __init__(self, cb, model_unique_id, initial_data=None):
        super(Notifier, self).__init__(cb, model_unique_id, initial_data)


class PendingAnalysis(MutableModel):
    urlobject = "/api/bit9platform/v1/pendingAnalysis"

    StatusScheduled = 0
    StatusSubmitted = 1
    StatusProcessed = 2
    StatusAnalyzed = 3
    StatusError = 4
    StatusCancelled = 5

    ResultNotAvailable = 0
    ResultClean = 1
    ResultPotentialThreat = 2
    ResultMalicious = 3

    def __init__(self, cb, model_unique_id, initial_data=None):
        super(PendingAnalysis, self).__init__(cb, model_unique_id, initial_data)

    @property
    def file(self):
        with closing(self._cb.session.get(self._build_api_request_uri() + "?downloadFile=true", stream=True)) as r:
            z = StringIO(r.content)
            zf = ZipFile(z)
            fp = zf.open(zf.filelist[0], "r")
            return fp

    def create_notification(self, **kwargs):
        n = self._cb.create(Notification, **kwargs)
        n.fileAnalysisId = self.id
        return n

    @property
    def fileCatalog(self):
        return self._cb.select(FileCatalog, self.fileCatalogId)

    @property
    def fileHash(self):
        return getattr(self, "md5", None) or getattr(self, "sha1", None) or getattr(self, "sha256", None)


class Policy(MutableBaseModel, CreatableModelMixin):
    urlobject = "/api/bit9platform/v1/policy"
    swagger_meta_file = "protection/models/policy.yaml"


class Publisher(MutableModel):
    urlobject = "/api/bit9platform/v1/publisher"

    def __init__(self, cb, model_unique_id, initial_data=None):
        super(Publisher, self).__init__(cb, model_unique_id, initial_data)


class PublisherCertificate(NewBaseModel):
    urlobject = "/api/bit9platform/v1/publisherCertificate"

    @classmethod
    def _minimum_server_version(cls):
        return LooseVersion("8.0")


class ScriptRule(MutableBaseModel):
    urlobject = "/api/bit9platform/v1/scriptRule"

    @classmethod
    def _minimum_server_version(cls):
        return LooseVersion("8.0")


@immutable
class ServerConfig(BaseModel):
    urlobject = "/api/bit9platform/v1/serverConfig"

    def __init__(self, cb, model_unique_id, initial_data=None):
        super(ServerConfig, self).__init__(cb, model_unique_id, initial_data)


@immutable
class ServerPerformance(BaseModel):
    urlobject = "/api/bit9platform/v1/serverPerformance"

    def __init__(self, cb, model_unique_id, initial_data=None):
        super(ServerPerformance, self).__init__(cb, model_unique_id, initial_data)


class Updater(MutableModel):
    urlobject = "/api/bit9platform/v1/updater"

    def __init__(self, cb, model_unique_id, initial_data=None):
        super(Updater, self).__init__(cb, model_unique_id, initial_data)


class TrustedDirectory(MutableBaseModel):
    urlobject = "/api/bit9platform/v1/trustedDirectory"

    @classmethod
    def _minimum_server_version(cls):
        return LooseVersion("8.0")


class TrustedUser(MutableBaseModel, CreatableModelMixin):
    urlobject = "/api/bit9platform/v1/trustedUser"
    swagger_meta_file = "protection/models/trustedUser.yaml"

    @classmethod
    def _minimum_server_version(cls):
        return LooseVersion("8.0")


class User(MutableBaseModel, CreatableModelMixin):
    urlobject = "/api/bit9platform/v1/user"
    swagger_meta_file = "protection/models/user.yaml"

    @classmethod
    def _minimum_server_version(cls):
        return LooseVersion("8.0")


class UserGroup(MutableBaseModel, CreatableModelMixin):
    urlobject = "/api/bit9platform/v1/userGroup"
    swagger_meta_file = "protection/models/userGroup.yaml"

    @classmethod
    def _minimum_server_version(cls):
        return LooseVersion("8.0")



#!/usr/bin/env python
#

import sys
from cbapi.response import Sensor
from cbapi.example_helpers import build_cli_parser, get_cb_response_object
import logging

log = logging.getLogger(__name__)


def download_file(sensor, path, output_filename=None):
    with sensor.lr_session() as session:

        # get basename of the file, if output_filename is None
        if output_filename is None:
            if session.os_type == 1:        # Windows uses backslashes
                output_filename = path.split('\\')[-1]
            else:
                output_filename = path.split('/')[-1]

        with open(output_filename, "wb") as fp:
            fp.write(session.get_file(path))
            print("Successfully wrote {0} to local file {1}.".format(path, output_filename))

    return 0


def main():
    parser = build_cli_parser("Download binary from endpoint through Live Response")
    parser.add_argument("-o", "--output", help="Output file name (default is the base file name)")
    parser.add_argument("-H", "--hostname", help="Hostname to download from", required=True)
    parser.add_argument("-p", "--path", help="Path to download", required=True)

    args = parser.parse_args()
    cb = get_cb_response_object(args)

    sensors = cb.select(Sensor).where("hostname:{0}".format(args.hostname))
    for sensor in sensors:
        if sensor.status == "Online":
            return download_file(sensor, args.path, args.output)

    print("No sensors for hostname {0} are online, exiting".format(args.hostname))


if __name__ == "__main__":
    sys.exit(main())



from cbapi.example_helpers import get_cb_protection_object, build_cli_parser
from cbapi.protection.models import Computer, Policy
import sys


def main():
    parser = build_cli_parser("Revert computers in policy to previous policy")
    parser.add_argument("--policy", "-p", help="Policy name or ID", required=True)
    args = parser.parse_args()

    p = get_cb_protection_object(args)

    try:
        policyId = int(args.policy)
    except ValueError:
        policyId = p.select(Policy).where("name:{}".format(args.policy)).id

    for computer in p.select(Computer).where("policyId:{}".format(policyId)):
        print("%s was in %s" % (computer.name, computer.policyName))
        computer.policyId = computer.previousPolicyId
        computer.save()
        print("%s is now in %s" % (computer.name, computer.policyName))


if __name__ == '__main__':
    sys.exit(main())



from __future__ import absolute_import

import json
import random
import string
import threading
import time
import logging
from collections import defaultdict

import shutil

from cbapi.errors import TimeoutError, ObjectNotFoundError, ApiError
from cbapi.six import itervalues
from concurrent.futures import _base, wait
from cbapi import winerror

from cbapi.six.moves.queue import Queue

from cbapi.response.models import Sensor


log = logging.getLogger(__name__)


class LiveResponseError(Exception):
    def __init__(self, details):
        message_list = []

        self.details = details
        self.win32_error = None
        self.decoded_win32_error = ""

        # Details object:
        # {u'status': u'error', u'username': u'admin', u'sensor_id': 9, u'name': u'kill',
        # u'completion': 1464319733.190924, u'object': 1660, u'session_id': 7, u'result_type': u'WinHresult',
        # u'create_time': 1464319733.171967, u'result_desc': u'', u'id': 22, u'result_code': 2147942487}

        if self.details.get("status") == "error" and self.details.get("result_type") == "WinHresult":
            # attempt to decode the win32 error
            win32_error_text = "Unknown Win32 error code"
            try:
                self.win32_error = int(self.details.get("result_code"))
                win32_error_text = "Win32 error code 0x%08X" % (self.win32_error,)
                self.decoded_win32_error = winerror.decode_hresult(self.win32_error)
                if self.decoded_win32_error:
                    win32_error_text += " ({0})".format(self.decoded_win32_error)
            except Exception:
                pass
            finally:
                message_list.append(win32_error_text)

        self.message = ": ".join(message_list)

    def __str__(self):
        return self.message


class CbLRSessionBase(object):
    MAX_RETRY_COUNT = 5

    def __init__(self, cblr_manager, session_id, sensor_id, session_data=None):
        self.session_id = session_id
        self.sensor_id = sensor_id
        self._cblr_manager = cblr_manager
        self._cb = cblr_manager._cb
        # TODO: refcount should be in a different object in the scheduler
        self._refcount = 1
        self._closed = False

        self.session_data = session_data
        self.os_type = None
        self.cblr_base = self._cblr_manager.cblr_base

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self._cblr_manager.close_session(self.sensor_id, self.session_id)
        self._closed = True

    def get_session_archive(self):
        response = self._cb.session.get("{cblr_base}/session/{0}/archive".format(self.session_id,
                                                                                 cblr_base=self.cblr_base), stream=True)
        response.raw.decode_content = True
        return response.raw

    #
    # File operations
    #
    def get_raw_file(self, file_name, timeout=None, delay=None):
        data = {"name": "get file", "object": file_name}

        resp = self._lr_post_command(data).json()
        file_id = resp.get('file_id', None)
        command_id = resp.get('id', None)

        self._poll_command(command_id, timeout=timeout, delay=delay)
        response = self._cb.session.get("{cblr_base}/session/{0}/file/{1}/content".format(self.session_id,
                                                                                          file_id,
                                                                                          cblr_base=self.cblr_base),
                                        stream=True)
        response.raw.decode_content = True
        return response.raw

    def get_file(self, file_name, timeout=None, delay=None):
        """
        Retrieve contents of the specified file name

        :param str file_name: Name of the file
        :return: Content of the specified file name
        :rtype: str
        """
        fp = self.get_raw_file(file_name, timeout=timeout, delay=delay)
        content = fp.read()
        fp.close()

        return content

    def delete_file(self, filename):
        """
        Delete the specified file name

        :param str filename: Name of the file
        :return: None
        """
        data = {"name": "delete file", "object": filename}
        resp = self._lr_post_command(data).json()
        command_id = resp.get('id')
        self._poll_command(command_id)

    def put_file(self, infp, remote_filename):
        """
        Create a new file on the remote endpoint with the specified data

        :Example:

        >>> with c.select(Sensor, 1).lr_session() as lr_session:
        ...     lr_session.put_file(open("test.txt", "rb"), r"c:\test.txt")

        :param str infp: Python file-like containing data to upload to the remote endpoint
        :param str remote_filename: File name to create on the remote endpoint
        :return: None
        """
        data = {"name": "put file", "object": remote_filename}
        file_id = self._upload_file(infp)
        data["file_id"] = file_id

        resp = self._lr_post_command(data).json()
        command_id = resp.get('id')
        self._poll_command(command_id)

    def list_directory(self, dir_name):
        """
        List the contents of a directory

        :Example:

        >>> with c.select(Sensor, 1).lr_session() as lr_session:
        ...     pprint.pprint(lr_session.list_directory('C:\\\\temp\\\\'))
        [{u'attributes': [u'DIRECTORY'],
          u'create_time': 1471897244,
          u'filename': u'.',
          u'last_access_time': 1476390670,
          u'last_write_time': 1476390670,
          u'size': 0},
         {u'attributes': [u'DIRECTORY'],
          u'create_time': 1471897244,
          u'filename': u'..',
          u'last_access_time': 1476390670,
          u'last_write_time': 1476390670,
          u'size': 0},
         {u'attributes': [u'ARCHIVE'],
          u'create_time': 1476390668,
          u'filename': u'test.txt',
          u'last_access_time': 1476390668,
          u'last_write_time': 1476390668,
          u'size': 0}]

        :param str dir_name: Directory to list.  This parameter should end with '\'
        :return: Returns a directory listing
        :rtype: list
        """
        data = {"name": "directory list", "object": dir_name}
        resp = self._lr_post_command(data).json()
        command_id = resp.get("id")
        return self._poll_command(command_id).get("files", [])

    def create_directory(self, dir_name):
        """
        Create a directory on the remote endpoint

        :param str dir_name: New directory name
        :return: None
        """
        data = {"name": "create directory", "object": dir_name}
        resp = self._lr_post_command(data).json()
        command_id = resp.get('id')
        self._poll_command(command_id)

    def path_join(self, *dirnames):
        if self.os_type == 1:
            # Windows
            return "\\".join(dirnames)
        else:
            # Unix/Mac OS X
            return "/".join(dirnames)

    def path_islink(self, fi):
        # TODO: implement
        return False

    def walk(self, top, topdown=True, onerror=None, followlinks=False):
        """
        Perform a full directory walk with recursion into subdirectories

        :Example:

        >>> with c.select(Sensor, 1).lr_session() as lr_session:
        ...     for entry in lr_session.walk(directory_name):
        ...         print(entry)
        ('C:\\temp\\', [u'dir1', u'dir2'], [u'file1.txt'])

        :param str top: Directory to recurse
        :param bool topdown: if True, start output from top level directory
        :param bool onerror: Callback if an error occurs.
                        This function is called with one argument (the exception that occurred)
        :param bool followlinks: Follow symbolic links
        :return: Returns output in the follow tuple format: (Directory Name, [dirnames], [filenames])
        :rtype: tuple
        """
        try:
            allfiles = self.list_directory(self.path_join(top, "*"))
        except Exception as err:
            if onerror is not None:
                onerror(err)
            return

        dirnames = []
        filenames = []

        for fn in allfiles:
            if "DIRECTORY" in fn["attributes"]:
                if fn["filename"] not in (".", ".."):
                    dirnames.append(fn)
            else:
                filenames.append(fn)

        if topdown:
            yield top, [fn["filename"] for fn in dirnames], [fn["filename"] for fn in filenames]

        for name in dirnames:
            new_path = self.path_join(top, name["filename"])
            if followlinks or not self.path_islink(new_path):
                for x in self.walk(new_path, topdown, onerror, followlinks):
                    yield x
        if not topdown:
            yield top, [fn["filename"] for fn in dirnames], [fn["filename"] for fn in filenames]

    #
    # Process operations
    #

    def kill_process(self, pid):
        """
        Terminate a process on the remote endpoint

        :param pid: Process ID to terminate
        :return: True if success, False if failure
        :rtype: bool
        """
        data = {"name": "kill", "object": pid}
        resp = self._lr_post_command(data).json()
        command_id = resp.get('id')

        try:
            self._poll_command(command_id, timeout=10, delay=0.1)
        except TimeoutError:
            return False

        return True

    def create_process(self, command_string, wait_for_output=True, remote_output_file_name=None,
                       working_directory=None, wait_timeout=30, wait_for_completion=True):
        """
        Create a new process with the specified command string.

        :Example:

        >>> with c.select(Sensor, 1).lr_session() as lr_session:
        ...     print(lr_session.create_process(r'cmd.exe /c "ping.exe 192.168.1.1"'))
        Pinging 192.168.1.1 with 32 bytes of data:
        Reply from 192.168.1.1: bytes=32 time<1ms TTL=64

        :param str command_string: command string used for the create process operation
        :param bool wait_for_output: Block on output from the new process (execute in foreground). This will also set
         wait_for_completion (below).
        :param str remote_output_file_name: The remote output file name used for process output
        :param str working_directory: The working directory of the create process operation
        :param int wait_timeout: Time out used for this live response command
        :param bool wait_for_completion: Wait until the process is completed before returning
        :return: returns the output of the command string
        :rtype: str
        """
        # process is:
        # - create a temporary file name
        # - create the process, writing output to a temporary file
        # - wait for the process to complete
        # - get the temporary file from the endpoint
        # - delete the temporary file

        if wait_for_output:
            wait_for_completion = True

        data = {"name": "create process", "object": command_string, "wait": wait_for_completion}

        if wait_for_output and not remote_output_file_name:
            randfilename = self._random_file_name()
            data["output_file"] = randfilename

        if working_directory:
            data["working_directory"] = working_directory

        if remote_output_file_name:
            data["output_file"] = remote_output_file_name

        resp = self._lr_post_command(data).json()
        command_id = resp.get('id')

        if wait_for_completion:
            self._poll_command(command_id, timeout=wait_timeout)

        if wait_for_output:
            # now the file is ready to be read

            file_content = self.get_file(data["output_file"])
            # delete the file
            self._lr_post_command({"name": "delete file", "object": data["output_file"]})

            return file_content
        else:
            return None

    def list_processes(self):
        """
        List currently running processes

        :Example:

        >>> with c.select(Sensor, 1).lr_session() as lr_session:
        ...     print(lr_session.list_processes()[0])
        {u'command_line': u'',
         u'create_time': 1476260500,
         u'parent': 0,
         u'parent_guid': u'00000001-0000-0000-0000-000000000000',
         u'path': u'',
         u'pid': 4,
         u'proc_guid': u'00000001-0000-0004-01d2-2461a85e4546',
         u'sid': u's-1-5-18',
         u'username': u'NT AUTHORITY\\SYSTEM'}

        :return: returns a list of running processes
        :rtype: list
        """
        data = {"name": "process list"}
        resp = self._lr_post_command(data).json()
        command_id = resp.get('id')

        return self._poll_command(command_id).get("processes", [])

    #
    # Registry operations
    #
    # returns dictionary with 2 entries ("values" and "sub_keys")
    #  "values" is a list containing a dictionary for each registry value in the key
    #  "sub_keys" is a list containing one entry for each sub_key
    def list_registry_keys_and_values(self, regkey):
        """
        Enumerate subkeys and values of the specified registry key.

        :Example:

        >>> with c.select(Sensor, 1).lr_session() as lr_session:
        >>>   pprint.pprint(lr_session.list_registry_keys_and_values('HKLM\\SYSTEM\\CurrentControlSet\\services\\ACPI'))
        {'sub_keys': [u'Parameters', u'Enum'],
         'values': [{u'value_data': 0,
                     u'value_name': u'Start',
                     u'value_type': u'REG_DWORD'},
                    {u'value_data': 1,
                     u'value_name': u'Type',
                     u'value_type': u'REG_DWORD'},
                    {u'value_data': 3,
                     u'value_name': u'ErrorControl',
                     u'value_type': u'REG_DWORD'},
                    {u'value_data': u'system32\\drivers\\ACPI.sys',
                     u'value_name': u'ImagePath',
                     u'value_type': u'REG_EXPAND_SZ'},
                    {u'value_data': u'Microsoft ACPI Driver',
                     u'value_name': u'DisplayName',
                     u'value_type': u'REG_SZ'},
                    {u'value_data': u'Boot Bus Extender',
                     u'value_name': u'Group',
                     u'value_type': u'REG_SZ'},
                    {u'value_data': u'acpi.inf_x86_neutral_ddd3c514822f1b21',
                     u'value_name': u'DriverPackageId',
                     u'value_type': u'REG_SZ'},
                    {u'value_data': 1,
                     u'value_name': u'Tag',
                     u'value_type': u'REG_DWORD'}]}

        :param str regkey: The registry key to enumerate
        :return: returns a dictionary with 2 keys (sub_keys and values)
        :rtype: dict
        """
        data = {"name": "reg enum key", "object": regkey}
        resp = self._lr_post_command(data).json()
        command_id = resp.get('id')
        results = {}
        results["values"] = self._poll_command(command_id).get("values", [])
        results["sub_keys"] = self._poll_command(command_id).get("sub_keys", [])
        return results

    # returns a list containing a dictionary for each registry value in the key
    def list_registry_keys(self, regkey):
        """
        Enumerate all registry values from the specified registry key.

        :param regkey: The registry key to enumearte
        :return: returns a list of values
        :rtype: list
        """
        data = {"name": "reg enum key", "object": regkey}
        resp = self._lr_post_command(data).json()
        command_id = resp.get('id')

        return self._poll_command(command_id).get("values", [])

    # returns a dictionary with the registry value
    def get_registry_value(self, regkey):
        """
        Returns the associated value of the specified registry key

        :Example:

        >>> with c.select(Sensor, 1).lr_session() as lr_session:
        >>>     pprint.pprint(lr_session.get_registry_value('HKLM\\SYSTEM\\CurrentControlSet\\services\\ACPI\\Start'))
        {u'value_data': 0, u'value_name': u'Start', u'value_type': u'REG_DWORD'}

        :param str regkey: The registry key to retrieve
        :return: Returns a dictionary with keys of: value_data, value_name, value_type
        :rtype: dict
        """
        data = {"name": "reg query value", "object": regkey}
        resp = self._lr_post_command(data).json()
        command_id = resp.get('id')

        return self._poll_command(command_id).get("value", {})

    def set_registry_value(self, regkey, value, overwrite=True, value_type=None):
        """
        Set a registry value of the specified registry key

        :Example:

        >>> with c.select(Sensor, 1).lr_session() as lr_session:
        ...     lr_session.set_registry_value('HKLM\\\\SYSTEM\\\\CurrentControlSet\\\\services\\\\ACPI\\\\testvalue', 1)

        :param str regkey: They registry key to set
        :param obj value: The value data
        :param bool overwrite: Overwrite value if True
        :param str value_type: The type of value.  Examples: REG_DWORD, REG_MULTI_SZ, REG_SZ
        :return: None
        """
        if value_type is None:
            if type(value) == int:
                value_type = "REG_DWORD"
            elif type(value) == list:
                value_type = "REG_MULTI_SZ"
            # elif type(value) == bytes:
            #     value_type = "REG_BINARY"
            else:
                value_type = "REG_SZ"
                value = str(value)

        data = {"name": "reg set value", "object": regkey, "overwrite": overwrite, "value_type": value_type,
                "value_data": value}
        resp = self._lr_post_command(data).json()
        command_id = resp.get('id')
        self._poll_command(command_id)

    def create_registry_key(self, regkey):
        """
        Create a new registry

        :param str regkey: The registry key to create
        :return: None
        """
        data = {"name": "reg create key", "object": regkey}
        resp = self._lr_post_command(data).json()
        command_id = resp.get('id')
        self._poll_command(command_id)

    def delete_registry_key(self, regkey):
        """
        Delete a registry key

        :param str regkey: The registry key to delete
        :return: None
        """
        data = {"name": "reg delete key", "object": regkey}
        resp = self._lr_post_command(data).json()
        command_id = resp.get('id')
        self._poll_command(command_id)

    def delete_registry_value(self, regkey):
        """
        Delete a registry value

        :param str regkey: the registry value to delete
        :return: None
        """
        data = {"name": "reg delete value", "object": regkey}
        resp = self._lr_post_command(data).json()
        command_id = resp.get('id')
        self._poll_command(command_id)

    #
    # Physical memory capture
    #
    def memdump(self, local_filename, remote_filename=None, compress=False):
        dump_object = self.start_memdump(remote_filename=remote_filename, compress=compress)
        dump_object.wait()
        dump_object.get(local_filename)
        dump_object.delete()

    def start_memdump(self, remote_filename=None, compress=True):
        if not remote_filename:
            remote_filename = self._random_file_name()

        data = {"name": "memdump", "object": remote_filename, "compress": compress}
        resp = self._lr_post_command(data).json()
        command_id = resp.get('id')

        if compress:
            remote_filename += ".zip"

        return LiveResponseMemdump(self, command_id, remote_filename)

    def _random_file_name(self):
        randfile = ''.join([random.choice(string.ascii_letters + string.digits) for _ in range(12)])
        if self.os_type == 1:
            workdir = 'c:\\windows\\temp'
        else:
            workdir = '/tmp'

        return self.path_join(workdir, 'cblr.%s.tmp' % (randfile,))

    def _poll_command(self, command_id, **kwargs):
        return poll_status(self._cb, "{cblr_base}/session/{0}/command/{1}".format(self.session_id, command_id,
                                                                                  cblr_base=self.cblr_base),
                           **kwargs)

    def _upload_file(self, fp):
        resp = self._cb.session.post("{cblr_base}/session/{0}/file".format(self.session_id, cblr_base=self.cblr_base),
                                     files={"file": fp}).json()
        return resp.get('id')

    def _lr_post_command(self, data):
        retries = self.MAX_RETRY_COUNT

        if "name" in data and data["name"] not in self.session_data["supported_commands"]:
            raise ApiError("Command {0} not supported by this sensor".format(data["name"]))

        while retries:
            try:
                data["session_id"] = self.session_id
                resp = self._cb.post_object("{cblr_base}/session/{0}/command".format(self.session_id,
                                                                                     cblr_base=self.cblr_base), data)
            except ObjectNotFoundError as e:
                if e.message.startswith("Sensor") or e.message.startswith("Session"):
                    self.session_id, self.session_data = self._cblr_manager._get_or_create_session(self.sensor_id)
                    retries -= 1
                    continue
                else:
                    try:
                        error_message = json.loads(e.message)
                        if error_message["status"] == "NOT_FOUND":
                            self.session_id, self.session_data = \
                                self._cblr_manager._get_or_create_session(self.sensor_id)
                            retries -= 1
                            continue
                    except Exception:
                        pass
                    raise ApiError("Received 404 error from server: {0}".format(e.message))
            else:
                return resp

        raise TimeoutError(message="Command {0} failed after {1} retries".format(data["name"], self.MAX_RETRY_COUNT))


class LiveResponseMemdump(object):
    def __init__(self, lr_session, memdump_id, remote_filename):
        self.lr_session = lr_session
        self.memdump_id = memdump_id
        self.remote_filename = remote_filename
        self._done = False
        self._error = None

    def get(self, local_filename):
        if not self._done:
            self.wait()
        if self._error:
            raise self._error
        src = self.lr_session.get_raw_file(self.remote_filename, timeout=3600, delay=5)
        dst = open(local_filename, "wb")
        shutil.copyfileobj(src, dst)

    def wait(self):
        self.lr_session._poll_command(self.memdump_id, timeout=3600, delay=5)
        self._done = True

    def delete(self):
        self.lr_session.delete_file(self.remote_filename)


def jobrunner(callable, cb, sensor_id):
    with cb.select(Sensor, sensor_id).lr_session() as sess:
        return callable(sess)


class WorkItem(object):
    def __init__(self, fn, sensor_id):
        self.fn = fn
        if isinstance(sensor_id, Sensor):
            self.sensor_id = sensor_id.id
        else:
            self.sensor_id = int(sensor_id)

        self.future = _base.Future()


class CompletionNotification(object):
    def __init__(self, sensor_id):
        self.sensor_id = sensor_id


class WorkerStatus(object):
    def __init__(self, sensor_id, status="ready", exception=None):
        self.sensor_id = sensor_id
        self.status = status
        self.exception = exception


class JobWorker(threading.Thread):
    def __init__(self, cb, sensor_id, result_queue):
        super(JobWorker, self).__init__()
        self.cb = cb
        self.sensor_id = sensor_id
        self.job_queue = Queue()
        self.lr_session = None
        self.result_queue = result_queue

    def run(self):
        try:
            self.lr_session = self.cb.live_response.request_session(self.sensor_id)
            self.result_queue.put(WorkerStatus(self.sensor_id, status="ready"))

            while True:
                work_item = self.job_queue.get(block=True)
                if not work_item:
                    self.job_queue.task_done()
                    return

                self.run_job(work_item)
                self.result_queue.put(CompletionNotification(self.sensor_id))
                self.job_queue.task_done()
        except Exception as e:
            self.result_queue.put(WorkerStatus(self.sensor_id, status="error", exception=e))
        finally:
            if self.lr_session:
                self.lr_session.close()
            self.result_queue.put(WorkerStatus(self.sensor_id, status="exiting"))

    def run_job(self, work_item):
        try:
            work_item.future.set_result(work_item.fn(self.lr_session))
        except Exception as e:
            work_item.future.set_exception(e)


class LiveResponseJobScheduler(threading.Thread):
    daemon = True

    def __init__(self, cb, max_workers=10):
        super(LiveResponseJobScheduler, self).__init__()
        self._cb = cb
        self._job_workers = {}
        self._idle_workers = set()
        self._unscheduled_jobs = defaultdict(list)
        self._max_workers = max_workers
        self.schedule_queue = Queue()

    def run(self):
        log.debug("Starting Live Response Job Scheduler")

        while True:
            log.debug("Waiting for item on Scheduler Queue")
            item = self.schedule_queue.get(block=True)
            log.debug("Got item: {0}".format(item))
            if isinstance(item, WorkItem):
                # new WorkItem available
                self._unscheduled_jobs[item.sensor_id].append(item)
            elif isinstance(item, CompletionNotification):
                # job completed
                self._idle_workers.add(item.sensor_id)
            elif isinstance(item, WorkerStatus):
                if item.status == "error":
                    log.error("Error encountered by JobWorker[{0}]: {1}".format(item.sensor_id,
                                                                                item.exception))
                elif item.status == "exiting":
                    log.debug("JobWorker[{0}] has exited, waiting...".format(item.sensor_id))
                    self._job_workers[item.sensor_id].join()
                    log.debug("JobWorker[{0}] deleted".format(item.sensor_id))
                    del self._job_workers[item.sensor_id]
                    try:
                        self._idle_workers.remove(item.sensor_id)
                    except KeyError:
                        pass
                elif item.status == "ready":
                    log.debug("JobWorker[{0}] now ready to accept jobs, session established".format(item.sensor_id))
                    self._idle_workers.add(item.sensor_id)
                else:
                    log.debug("Unknown status from JobWorker[{0}]: {1}".format(item.sensor_id, item.status))
            else:
                log.debug("Received unknown item on the scheduler Queue, exiting")
                # exiting the scheduler if we get None
                # TODO: wait for all worker threads to exit
                return

            self._schedule_jobs()

    def _schedule_jobs(self):
        log.debug("Entering scheduler")

        # First, see if there are new jobs to schedule on idle workers.
        self._schedule_existing_workers()

        # If we have jobs scheduled to run on sensors with no current associated worker, let's spawn new ones.
        if set(self._unscheduled_jobs.keys()) - self._idle_workers:
            self._cleanup_idle_workers()
            self._spawn_new_workers()
            self._schedule_existing_workers()

    def _cleanup_idle_workers(self, max=None):
        if not max:
            max = self._max_workers

        for sensor in list(self._idle_workers)[:max]:
            log.debug("asking worker for sensor id {0} to exit".format(sensor))
            self._job_workers[sensor].job_queue.put(None)

    def _schedule_existing_workers(self):
        log.debug("There are idle workers for sensor ids {0}".format(self._idle_workers))

        intersection = self._idle_workers.intersection(set(self._unscheduled_jobs.keys()))

        log.debug("{0} jobs ready to execute in existing execution slots".format(len(intersection)))

        for sensor in intersection:
            item = self._unscheduled_jobs[sensor].pop(0)
            self._job_workers[sensor].job_queue.put(item)
            self._idle_workers.remove(item.sensor_id)

        self._cleanup_unscheduled_jobs()

    def _cleanup_unscheduled_jobs(self):
        marked_for_deletion = []
        for k in self._unscheduled_jobs.keys():
            if len(self._unscheduled_jobs[k]) == 0:
                marked_for_deletion.append(k)

        for k in marked_for_deletion:
            del self._unscheduled_jobs[k]

    def submit_job(self, work_item):
        self.schedule_queue.put(work_item)

    def _spawn_new_workers(self):
        if len(self._job_workers) >= self._max_workers:
            return

        schedule_max = self._max_workers - len(self._job_workers)

        sensors = [s for s in self._cb.select(Sensor) if s.id in self._unscheduled_jobs
                   and s.id not in self._job_workers
                   and s.status == "Online"]
        sensors_to_schedule = sorted(sensors, key=lambda x: x.next_checkin_time)[:schedule_max]

        log.debug("Spawning new workers to handle these sensors: {0}".format(sensors_to_schedule))
        for sensor in sensors_to_schedule:
            log.debug("Spawning new JobWorker for sensor id {0}".format(sensor.id))
            self._job_workers[sensor.id] = JobWorker(self._cb, sensor.id, self.schedule_queue)
            self._job_workers[sensor.id].start()


class CbLRManagerBase(object):
    cblr_base = ""                        # override in subclass for each product
    cblr_session_cls = NotImplemented     # override in subclass for each product

    def __init__(self, cb, timeout=30, keepalive_sessions=False):
        self._timeout = timeout
        self._cb = cb
        self._sessions = {}
        self._session_lock = threading.RLock()
        self._keepalive_sessions = keepalive_sessions

        if keepalive_sessions:
            self._cleanup_thread = threading.Thread(target=self._session_keepalive_thread)
            self._cleanup_thread.daemon = True
            self._cleanup_thread.start()

        self._job_scheduler = None

    def submit_job(self, job, sensor):
        if self._job_scheduler is None:
            # spawn the scheduler thread
            self._job_scheduler = LiveResponseJobScheduler(self._cb)
            self._job_scheduler.start()

        work_item = WorkItem(job, sensor)
        self._job_scheduler.submit_job(work_item)
        return work_item.future

    def _session_keepalive_thread(self):
        log.debug("Starting Live Response scheduler cleanup task")
        while True:
            time.sleep(self._timeout)

            delete_list = []
            with self._session_lock:
                for session in itervalues(self._sessions):
                    if session._refcount == 0:
                        delete_list.append(session.sensor_id)
                    else:
                        try:
                            self._send_keepalive(session.session_id)
                        except ObjectNotFoundError:
                            log.debug("Session {0} for sensor {1} not valid any longer, removing from cache"
                                      .format(session.session_id, session.sensor_id))
                            delete_list.append(session.sensor_id)
                        except Exception:
                            log.debug(("Keepalive on session {0} (sensor {1}) failed with unknown error, " +
                                      "removing from cache").format(session.session_id, session.sensor_id))
                            delete_list.append(session.sensor_id)

                for sensor_id in delete_list:
                    self._close_session(self._sessions[sensor_id].session_id)
                    del self._sessions[sensor_id]

    def request_session(self, sensor_id):
        if self._keepalive_sessions:
            with self._session_lock:
                if sensor_id in self._sessions:
                    session = self._sessions[sensor_id]
                    self._sessions[sensor_id]._refcount += 1
                else:
                    session_id, session_data = self._get_or_create_session(sensor_id)
                    session = self.cblr_session_cls(self, session_id, sensor_id, session_data=session_data)
                    self._sessions[sensor_id] = session
        else:
            session_id, session_data = self._get_or_create_session(sensor_id)
            session = self.cblr_session_cls(self, session_id, sensor_id, session_data=session_data)

        return session

    def close_session(self, sensor_id, session_id):
        if self._keepalive_sessions:
            with self._session_lock:
                try:
                    self._sessions[sensor_id]._refcount -= 1
                except KeyError:
                    pass
        else:
            self._close_session(session_id)

    def _send_keepalive(self, session_id):
        log.debug("Sending keepalive message for session id {0}".format(session_id))
        self._cb.get_object("{cblr_base}/session/{0}/keepalive".format(session_id, cblr_base=self.cblr_base))


class GetFileJob(object):
    def __init__(self, file_name):
        self._file_name = file_name

    def run(self, session):
        return session.get_file(self._file_name)


# TODO: adjust the polling interval and also provide a callback function to report progress
def poll_status(cb, url, desired_status="complete", timeout=None, delay=None):
    start_time = time.time()
    status = None

    if not timeout:
        timeout = 120
    if not delay:
        delay = 0.5

    while status != desired_status and time.time() - start_time < timeout:
        res = cb.get_object(url)
        if res["status"] == desired_status:
            log.debug(json.dumps(res))
            return res
        elif res["status"] == "error":
            raise LiveResponseError(res)
        else:
            time.sleep(delay)

    raise TimeoutError(uri=url, message="timeout polling for Live Response")


if __name__ == "__main__":
    from cbapi.response import CbEnterpriseResponseAPI
    import logging
    root = logging.getLogger()
    root.addHandler(logging.StreamHandler())

    logging.getLogger("cbapi").setLevel(logging.DEBUG)

    c = CbEnterpriseResponseAPI()
    j = GetFileJob(r"c:\test.txt")
    with c.select(Sensor, 3).lr_session() as lr_session:
        file_contents = lr_session.get_file(r"c:\test.txt")

    future = c.live_response.submit_job(j.run, 3)
    wait([future, ])
    print(future.result())



#!/usr/bin/env python
#

import sys
from cbapi.response.models import Feed
from cbapi.example_helpers import build_cli_parser, get_cb_response_object, get_object_by_name_or_id
from cbapi.errors import ServerError
import logging

log = logging.getLogger(__name__)


def list_feeds(cb, parser, args):
    for f in cb.select(Feed):
        for fieldname in ["id", "category", "display_name", "enabled", "provider_url", "summary", "tech_data",
                          "feed_url", "use_proxy", "validate_server_cert"]:
            print("%-20s: %s" % (fieldname, getattr(f, fieldname, "")))

        if f.username:
            for fieldname in ["username", "password"]:
                print("%-20s: %s" % (fieldname, getattr(f, fieldname, "")))

        if f.ssl_client_crt:
            for fieldname in ["ssl_client_crt", "ssl_client_key"]:
                print("%-20s: %s" % (fieldname, getattr(f, fieldname, "")))

        print("\n")


def add_feed(cb, parser, args):
    configured_feeds = [f for f in cb.select(Feed) if f.feed_url == args.feed_url]
    if len(configured_feeds):
        print("Warning: Feeds already configured for this url: {0:s}:".format(args.feed_url))
        for f in configured_feeds:
            print(f)
            print("")
        if not args.force:
            return

    f = cb.create(Feed)
    f.feed_url = args.feed_url
    if args.enable:
        f.enabled = True

    if args.username:
        f.username = args.username
    if args.password:
        f.password = args.password

    if args.cert:
        f.ssl_client_crt = open(args.cert, "rb").read()
    if args.key:
        f.ssl_client_key = open(args.key, "rb").read()

    if args.use_proxy:
        f.use_proxy = True

    if args.validate_server_cert:
        f.validate_server_cert = True

    log.debug("Adding feed: {0:s}".format(str(f)))

    try:
        f.save()
    except ServerError as se:
        if se.error_code == 500:
            print("Could not add feed:")
            print(" Received error code 500 from server. This is usually because the server cannot retrieve the feed.")
            print(" Check to ensure the Cb server has network connectivity and the credentials are correct.")
        else:
            print("Could not add feed: {0:s}".format(str(se)))
    except Exception as e:
        print("Could not add feed: {0:s}".format(str(e)))
    else:
        log.debug("Feed data: {0:s}".format(str(f)))
        print("Added feed. New feed ID is {0:d}".format(f.id))


def delete_feed(cb, parser, args):
    try:
        if args.id:
            attempted_to_find = "ID of {0:d}".format(args.id)
            feeds = [cb.select(Feed, args.id, force_init=True)]
        else:
            attempted_to_find = "name {0:s}".format(args.feedname)
            feeds = cb.select(Feed).where("name:{0:s}".format(args.feedname))[::]
            if not len(feeds):
                raise Exception("No feeds match")
    except Exception as e:
        print("Could not find feed with {0:s}: {1:s}".format(attempted_to_find, str(e)))
        return

    num_matching_feeds = len(feeds)
    if num_matching_feeds > 1 and not args.force:
        print("{0:d} feeds match {1:s} and --force not specified. No action taken.".format(num_matching_feeds,
                                                                                           attempted_to_find))
        return

    for f in feeds:
        try:
            f.delete()
        except Exception as e:
            print("Could not delete feed with {0:s}: {1:s}".format(attempted_to_find, str(e)))
        else:
            print("Deleted feed id {0:d} with name {1:s}".format(f.id, f.name))


def toggle_feed(cb, feedname, enable=True):
    try:
        feeds = list(cb.select(Feed).where("name:{0:s}".format(feedname)))
    except Exception as e:
        print("Could not find any feeds with the name {0:s}: {1:s}".format(feedname, str(e)))
        return

    for feed in feeds:
        if enable:
            operation = "Enabl"
        else:
            operation = "Disabl"

        if feed.enabled == enable:
            print("Feed named {0} with id {1} already {2}ed, continuing".format(feed.name, feed.id, operation))
        else:
            print("{0}ing feed named {1:s} with id {2}".format(operation, feed.name, feed.id))
            feed.enabled = enable
            try:
                feed.save()
            except Exception as e:
                print("-> Encountered error {0}ing feed id {1}: {2}".format(operation, feed.id, str(e)))


def list_actions(cb, parser, args):
    feeds = get_object_by_name_or_id(cb, Feed, name=args.feedname, id=args.id)
    if len(feeds) > 1:
        print("Multiple feeds match the name {}, giving up.".format(args.feedname))
        return

    feed = feeds[0]
    print("Actions in Feed {}:".format(feed.name))
    for action in feed.actions:
        print("  - {0}: {1}".format(action.type, action.action_data))


def main():
    parser = build_cli_parser()
    commands = parser.add_subparsers(help="Feed commands", dest="command_name")

    commands.add_parser("list", help="List all configured feeds")

    list_actions_command = commands.add_parser("list-actions", help="List actions associated with a feed")
    list_actions_specifier = list_actions_command.add_mutually_exclusive_group(required=True)
    list_actions_specifier.add_argument("-i", "--id", type=int, help="ID of feed")
    list_actions_specifier.add_argument("-f", "--feedname", help="Name of feed")

    add_command = commands.add_parser("add", help="Add new feed")
    add_command.add_argument("-u", "--feed-url", help="URL location of feed data", dest="feed_url", required=True)
    add_command.add_argument("--force", help="Force creation even if feed already exists", action="store_true")
    add_command.add_argument("-e", "--enable", action="store_true", help="Enable this feed")
    add_command.add_argument("-p", "--use_proxy", action="store_true", default=False, dest="use_proxy",
                             help="Carbon Black server will use configured web proxy to download feed from feed url")
    add_command.add_argument("-v", "--validate_server_cert", action="store_true", default=False,
                             dest="validate_server_cert",
                             help="Carbon Black server will verify the SSL certificate of the feed server")
    http_basic_auth = add_command.add_argument_group("HTTP Authentication")
    http_basic_auth.add_argument("--username", help="HTTP Basic Authentication username to access Feed URL")
    http_basic_auth.add_argument("--password", help="HTTP Basic Authentication password to access Feed URL")
    tls_auth = add_command.add_argument_group("TLS Client Authentication")
    tls_auth.add_argument("--cert",
                          help="Path to file containing TLS client certificate required to access Feed URL")
    tls_auth.add_argument("--key", help="Path to file containing TLS client key required to access Feed URL")

    del_command = commands.add_parser("delete", help="Delete feeds")
    del_feed_specifier = del_command.add_mutually_exclusive_group(required=True)
    del_feed_specifier.add_argument("-i", "--id", type=int, help="ID of feed to delete")
    del_feed_specifier.add_argument("-f", "--feedname", help="Name of feed to delete. Specify --force to delete"
                                    " multiple feeds that have the same name")
    del_command.add_argument("--force", help="If FEEDNAME matches multiple feeds, delete all matching feeds",
                             action="store_true", default=False)

    enable_command = commands.add_parser("enable", help="Enable a feed")
    enable_command.add_argument("-f", "--feedname", help="Name of feed to enable", required=True)

    disable_command = commands.add_parser("disable", help="Disable a feed")
    disable_command.add_argument("-f", "--feedname", help="Name of feed to disable", required=True)

    args = parser.parse_args()
    cb = get_cb_response_object(args)

    if args.command_name == "list":
        return list_feeds(cb, parser, args)
    elif args.command_name == "list-actions":
        return list_actions(cb, parser, args)
    elif args.command_name == "add":
        return add_feed(cb, parser, args)
    elif args.command_name == "delete":
        return delete_feed(cb, parser, args)
    elif args.command_name in ("disable", "enable"):
        return toggle_feed(cb, args.feedname, enable=(args.command_name == "enable"))


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

import sys
from cbapi.example_helpers import build_cli_parser, get_cb_psc_object
from cbapi.psc.models import BaseAlert
from helpers.alertsv6 import setup_parser_with_basic_criteria, load_basic_criteria


def main():
    parser = build_cli_parser("List alert facets")
    setup_parser_with_basic_criteria(parser)
    parser.add_argument("-F", "--facet", action="append", choices=["ALERT_TYPE", "CATEGORY", "REPUTATION", "WORKFLOW",
                                                                   "TAG", "POLICY_ID", "POLICY_NAME", "DEVICE_ID",
                                                                   "DEVICE_NAME", "APPLICATION_HASH",
                                                                   "APPLICATION_NAME", "STATUS", "RUN_STATE",
                                                                   "POLICY_APPLIED_STATE", "POLICY_APPLIED",
                                                                   "SENSOR_ACTION"],
                        required=True, help="Retrieve these fields as facet information")

    args = parser.parse_args()
    cb = get_cb_psc_object(args)

    query = cb.select(BaseAlert)
    load_basic_criteria(query, args)

    facetinfos = query.facets(args.facet)
    for facetinfo in facetinfos:
        print("For field '{0}':".format(facetinfo["field"]))
        for facetval in facetinfo["values"]:
            print("\tValue {0}: {1} occurrences".format(facetval["id"], facetval["total"]))


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/python

from cbapi.six import iteritems
from cbapi.six.moves.configparser import RawConfigParser
from cbapi.protection import Connector, Notification, PendingAnalysis
from cbapi.example_helpers import get_cb_protection_object, build_cli_parser
import requests
import tempfile
import shutil
import datetime
import logging
import time
import sys

# logging
logging.basicConfig(format='%(asctime)s %(levelname)s:%(message)s', level=logging.DEBUG)
logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)

log = logging.getLogger(__name__)


class VirusTotalConnector(object):
    def __init__(self, api, vt_token=None, connector_name='VirusTotal', allow_uploads=True, malicious_threshold=50,
                 potential_threshold=10):
        """VirusTotal connector main object.

        :param cbapi.protection.CbEnterpriseResponseAPI: api: API object
        :param str vt_token: API token provided by VirusTotal
        :param str connector_name: name of the connector. Defaults to 'VirusTotal'
        :param bool allow_uploads: True to allow uploads of binaries to the VirusTotal servers. If set to False,
            only hash lookups will be done to te virusTotal.
            Note: In case when allow_uploads is set to False  AND VirusTotal does not recognize the hash,
            associated Cb Protection file analysis request will be cancelled
        """

        if vt_token is None:
            raise TypeError("Missing required VT authentication token.")
        self.vt_token = vt_token
        self.vt_url = 'https://www.virustotal.com/vtapi/v2'
        self.polling_frequency = 30  # seconds

        self.malicious_threshold = malicious_threshold
        self.potential_threshold = potential_threshold

        # Global dictionary to track our VT scheduled scans. We need this since it takes VT a while to process results
        # and we don't want to keep polling VT too often
        # Any pending results will be kept here, together with next polling time
        self.awaiting_results = {}

        self.allow_uploads = allow_uploads

        self.connector = api.create(Connector, data={"name": connector_name, "analysisName": connector_name,
                                                     "connectorVersion": "1.0", "canAnalyze": True,
                                                     "analysisEnabled": True, "enabled": True})
        self.connector.save()

        log.info("Connector: {0:s}".format(str(self.connector)))

    def run(self):
        while True:
            for binary in self.connector.pendingAnalyses:
                self.process_request(binary)
            time.sleep(self.polling_frequency)

    def check_vt(self, resource_id):
        r = requests.get("{0:s}/file/report".format(self.vt_url),
                         params={"resource": resource_id, "apikey": self.vt_token})

        if r.status_code == 204:
            log.info("VirusTotal API rate limit reached, sleeping for 30 seconds")
            time.sleep(30)

        return r

    def process_response(self, binary, scanResults):
        if "positives" in scanResults:
            self.report_result(binary, scanResults)
        elif "scanId" in scanResults:
            self.schedule_check(binary, scanResults["scanId"])
        elif self.allow_uploads:
            self.upload_to_vt(binary)
            self.schedule_check(binary, binary.fileHash)
        else:
            binary.analysisStatus = PendingAnalysis.StatusCancelled
            log.info("%s: VirusTotal has no information and we aren't allowed to upload it. "
                     "Cancelling the analysis request." % binary.fileHash)
            binary.save()

    def report_result(self, binary, scanResults):
        # We have results. Create our notification
        n = binary.create_notification(data={"product": "VirusTotal", "malwareName": "", "malwareType": ""})

        # Let's see if it is malicious. Use some fancy heuristics...
        positivesPerc = 100 * scanResults.get('positives') / scanResults.get('total')
        if positivesPerc > self.malicious_threshold:
            n.analysisResult = Notification.ResultMalicious
            n.severity = "critical"
            n.type = "malicious_file"
        elif positivesPerc > self.potential_threshold:
            n.analysisResult = Notification.ResultPotentialThreat
            n.severity = "high"
            n.type = "potential_risk_file"
        else:
            n.analysisResult = Notification.ResultClean
            n.severity = "low"
            n.type = "clean_file"

        n.externalUrl = scanResults.get('permalink')

        # Enumerate scan results that have detected the issue and build our
        # 'malwareName' string for the notification
        scans = scanResults.get("scans", {})
        malware_type = [k + ":" + v["result"] for k, v in iteritems(scans) if v["detected"]]
        malware_name = [v["result"] for k, v in iteritems(scans) if v["detected"]]

        n.malwareType = "; ".join(malware_type[:4])
        n.malwareName = "; ".join(malware_name[:4])

        if len(malware_type) > 4:
            n.malwareName += "..."
            n.malwareType += "..."

        # Send notification
        n.save()

        if binary.fileHash in self.awaiting_results:
            del self.awaiting_results[binary.fileHash]

        log.info("VT analysis for %s completed. VT result is %d%% malware (%s). Reporting status: %s"
                 % (binary.fileHash, positivesPerc, n.malwareName, n.type))

    def upload_to_vt(self, binary):
        if binary.uploaded:
            log.info("%s: VirusTotal has no information on this hash. Uploading the file" % binary.fileHash)

            vt_upload_error = False

            with tempfile.NamedTemporaryFile() as outfp:
                shutil.copyfileobj(binary.file, outfp)
                outfp.seek(0)
                files = {'file': outfp}
                try:
                    r = requests.post(self.vt_url + "/file/scan", files=files, params={'apikey': self.vt_token})
                    if r.status_code != 200:
                        vt_upload_error = True
                except Exception:
                    log.exception("Could not send file %s to VirusTotal" % (binary.fileHash,))
                    vt_upload_error = True

            if vt_upload_error:
                binary.analysisStatus = PendingAnalysis.StatusError
                binary.analysisError = 'VirusTotal returned error when attempting to send file for scanning'
            else:
                binary.analysisStatus = PendingAnalysis.StatusSubmitted

            binary.save()
        else:
            log.info("%s: VirusTotal has no information on this hash. Waiting for agent to upload it."
                     % binary.fileHash)

    def schedule_check(self, binary, scanId):
        next_check = datetime.datetime.now() + datetime.timedelta(0, 3600)
        self.awaiting_results[binary.fileHash] = {'scanId': scanId, 'nextCheck': next_check}
        log.info("%s: Waiting for analysis to complete. Will check back after %s."
                 % (binary.fileHash, next_check.strftime("%Y-%m-%d %H:%M:%S")))

    def process_request(self, binary):
        if binary.fileHash in self.awaiting_results:
            lastAttempt = self.awaiting_results[binary.fileHash]
            if lastAttempt["nextCheck"] > datetime.datetime.now():
                return

            scanId = lastAttempt["scanId"]
            r = self.check_vt(scanId)
        else:
            r = self.check_vt(binary.fileHash)

        if r.status_code != 200:
            return

        scanResults = r.json()

        self.process_response(binary, scanResults)


def main():
    parser = build_cli_parser("VirusTotal Connector")
    parser.add_argument("--config", "-c", help="Path to configuration file", default="virustotal.ini")
    args = parser.parse_args()

    inifile = RawConfigParser({
        "vt_api_key": None,
        "retrieve_files": "true",
        "upload_binaries_to_vt": "false",
        "connector_name": "VirusTotal",
        "log_file": None,
    })
    inifile.read(args.config)

    config = {}
    config["vt_api_key"] = inifile.get("bridge", "vt_api_key")
    config["retrieve_files"] = inifile.getboolean("bridge", "retrieve_files")
    config["connector_name"] = inifile.get("bridge", "connector_name")
    config["upload_binaries_to_vt"] = inifile.getboolean("bridge", "upload_binaries_to_vt")

    log_file = inifile.get("bridge", "log_file")
    if log_file:
        file_handler = logging.FileHandler(log_file)
        formatter = logging.Formatter('%(asctime)s %(levelname)s:%(message)s')
        file_handler.setFormatter(formatter)
        file_handler.setLevel(logging.DEBUG)
        logging.getLogger().addHandler(file_handler)

    if not config["vt_api_key"]:
        log.fatal("Cannot start without a valid VirusTotal API key, exiting")
        return 1

    log.info("Configuration:")
    for k, v in iteritems(config):
        log.info("    %-20s: %s" % (k, v))

    api = get_cb_protection_object(args)

    vt = VirusTotalConnector(
        api,
        vt_token=config["vt_api_key"],
        allow_uploads=config["upload_binaries_to_vt"],  # Allow VT connector to upload binary files to VirusTotal
        connector_name=config["connector_name"],
    )

    log.info("Starting VirusTotal processing loop")
    vt.run()


if __name__ == '__main__':
    sys.exit(main())
